@@ -9,16 +9,18 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 import unittest
-import xml.etree.ElementTree as ElementTree
+from xml.etree.ElementTree import Element, parse as etree_parse
 from pathlib import Path
+from typing import cast, MutableMapping, Optional, Type
 
 try:
     import lxml.etree as lxml_etree
 except ImportError:
     lxml_etree = None
 
+from elementpath.etree import etree_tostring
+
 from xmlschema import XMLSchema, XMLSchemaValidationError, fetch_namespaces
-from xmlschema.etree import etree_element
 from xmlschema.dataobjects import DataElement
 from xmlschema.testing import etree_elements_assert_equal
 
@@ -29,29 +31,43 @@ from xmlschema.dataobjects import DataElementConverter
 
 
 class TestConverters(unittest.TestCase):
+    col_xsd_filename: str
+    col_xml_filename: str
+    col_nsmap: MutableMapping[str, str]
+    col_lxml_root: Optional['lxml_etree.ElementTree']
+
+    col_xsd_filename: str
+    col_xml_filename: str
+    col_nsmap: dict
 
     @classmethod
     def setUpClass(cls):
         cls.col_xsd_filename = cls.casepath('examples/collection/collection.xsd')
         cls.col_xml_filename = cls.casepath('examples/collection/collection.xml')
-        cls.col_xml_root = ElementTree.parse(cls.col_xml_filename).getroot()
-        cls.col_lxml_root = lxml_etree.parse(cls.col_xml_filename).getroot()
+        cls.col_xml_root = etree_parse(cls.col_xml_filename).getroot()
         cls.col_nsmap = fetch_namespaces(cls.col_xml_filename)
         cls.col_namespace = cls.col_nsmap['col']
 
+        if lxml_etree is not None:
+            cls.col_lxml_root = lxml_etree.parse(cls.col_xml_filename).getroot()
+        else:
+            cls.col_lxml_root = None
+
     @classmethod
     def casepath(cls, relative_path):
         return str(Path(__file__).parent.joinpath('test_cases', relative_path))
 
     def test_element_class_argument(self):
         converter = XMLSchemaConverter()
-        self.assertIs(converter.etree_element_class, etree_element)
+        self.assertIs(converter.etree_element_class, Element)
 
-        converter = XMLSchemaConverter(etree_element_class=etree_element)
-        self.assertIs(converter.etree_element_class, etree_element)
+        converter = XMLSchemaConverter(etree_element_class=Element)
+        self.assertIs(converter.etree_element_class, Element)
 
         if lxml_etree is not None:
-            converter = XMLSchemaConverter(etree_element_class=lxml_etree.Element)
+            converter = XMLSchemaConverter(
+                etree_element_class=cast(Type[Element], lxml_etree.Element)
+            )
             self.assertIs(converter.etree_element_class, lxml_etree.Element)
 
     def test_prefix_arguments(self):
@@ -94,7 +110,7 @@ class TestConverters(unittest.TestCase):
                     </xs:sequence>
                 </xs:complexType>
             </xs:element>
-        </xs:schema> 
+        </xs:schema>
         """)
 
         self.assertEqual(
@@ -107,7 +123,7 @@ class TestConverters(unittest.TestCase):
 
     def test_preserve_root__issue_215(self):
         schema = XMLSchema("""
-        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" 
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
                    xmlns="http://xmlschema.test/ns"
                    targetNamespace="http://xmlschema.test/ns">
             <xs:element name="a">
@@ -118,8 +134,7 @@ class TestConverters(unittest.TestCase):
                     </xs:sequence>
                 </xs:complexType>
             </xs:element>
-        </xs:schema> 
-        """)
+        </xs:schema>""")
 
         xml_data = """<tns:a xmlns:tns="http://xmlschema.test/ns"><b1/><b2/></tns:a>"""
 
@@ -143,10 +158,10 @@ class TestConverters(unittest.TestCase):
     def test_etree_element_method(self):
         converter = XMLSchemaConverter()
         elem = converter.etree_element('A')
-        self.assertIsNone(etree_elements_assert_equal(elem, etree_element('A')))
+        self.assertIsNone(etree_elements_assert_equal(elem, Element('A')))
 
         elem = converter.etree_element('A', attrib={})
-        self.assertIsNone(etree_elements_assert_equal(elem, etree_element('A')))
+        self.assertIsNone(etree_elements_assert_equal(elem, Element('A')))
 
     def test_columnar_converter(self):
         col_schema = XMLSchema(self.col_xsd_filename, converter=ColumnarConverter)
@@ -196,9 +211,10 @@ class TestConverters(unittest.TestCase):
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
 
         # Decode from lxml.etree.Element tree
-        obj2 = col_schema.decode(self.col_lxml_root)
-        self.assertIn("'@xmlns:col'", repr(obj2))
-        self.assertEqual(obj1, obj2)
+        if self.col_lxml_root is not None:
+            obj2 = col_schema.decode(self.col_lxml_root)
+            self.assertIn("'@xmlns:col'", repr(obj2))
+            self.assertEqual(obj1, obj2)
 
         # Decode from ElementTree.Element tree providing namespaces
         obj2 = col_schema.decode(self.col_xml_root, namespaces=self.col_nsmap)
@@ -232,10 +248,11 @@ class TestConverters(unittest.TestCase):
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
 
         # Decode from lxml.etree.Element tree
-        obj2 = col_schema.decode(self.col_lxml_root, preserve_root=True)
-        self.assertIn("'col:collection'", repr(obj2))
-        self.assertIn("'@xmlns:col'", repr(obj2))
-        self.assertEqual(obj1, obj2)
+        if self.col_lxml_root is not None:
+            obj2 = col_schema.decode(self.col_lxml_root, preserve_root=True)
+            self.assertIn("'col:collection'", repr(obj2))
+            self.assertIn("'@xmlns:col'", repr(obj2))
+            self.assertEqual(obj1, obj2)
 
         # Decode from ElementTree.Element tree providing namespaces
         obj2 = col_schema.decode(self.col_xml_root, namespaces=self.col_nsmap, preserve_root=True)
@@ -270,9 +287,10 @@ class TestConverters(unittest.TestCase):
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
 
         # Decode from lxml.etree.Element tree
-        obj2 = col_schema.decode(self.col_lxml_root)
-        self.assertIn("'@xmlns:col'", repr(obj2))
-        self.assertEqual(obj1, obj2)
+        if self.col_lxml_root is not None:
+            obj2 = col_schema.decode(self.col_lxml_root)
+            self.assertIn("'@xmlns:col'", repr(obj2))
+            self.assertEqual(obj1, obj2)
 
         # Decode from ElementTree.Element tree providing namespaces
         obj2 = col_schema.decode(self.col_xml_root, namespaces=self.col_nsmap)
@@ -306,10 +324,11 @@ class TestConverters(unittest.TestCase):
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
 
         # Decode from lxml.etree.Element tree
-        obj2 = col_schema.decode(self.col_lxml_root, preserve_root=True)
-        self.assertIn("'col:collection'", repr(obj2))
-        self.assertIn("'@xmlns:col'", repr(obj2))
-        self.assertEqual(obj1, obj2)
+        if self.col_lxml_root is not None:
+            obj2 = col_schema.decode(self.col_lxml_root, preserve_root=True)
+            self.assertIn("'col:collection'", repr(obj2))
+            self.assertIn("'@xmlns:col'", repr(obj2))
+            self.assertEqual(obj1, obj2)
 
         # Decode from ElementTree.Element tree providing namespaces
         obj2 = col_schema.decode(self.col_xml_root, namespaces=self.col_nsmap, preserve_root=True)
@@ -339,7 +358,7 @@ class TestConverters(unittest.TestCase):
             col_schema.encode(obj1, path='./col:collection', namespaces=self.col_nsmap)
         self.assertIn("missing required attribute 'id'", str(ec.exception))
 
-    def test_decode_encode_badger_fish_converter(self):
+    def test_decode_encode_badgerfish_converter(self):
         col_schema = XMLSchema(self.col_xsd_filename, converter=BadgerFishConverter)
 
         obj1 = col_schema.decode(self.col_xml_filename)
@@ -358,7 +377,8 @@ class TestConverters(unittest.TestCase):
         self.assertEqual(obj1, col_schema.decode(self.col_xml_root, namespaces=self.col_nsmap))
 
         # With lxml.etree namespaces are mapped
-        self.assertEqual(obj1, col_schema.decode(self.col_lxml_root))
+        if self.col_lxml_root is not None:
+            self.assertEqual(obj1, col_schema.decode(self.col_lxml_root))
 
         root = col_schema.encode(obj2, path='./col:collection', namespaces=self.col_nsmap)
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
@@ -385,7 +405,8 @@ class TestConverters(unittest.TestCase):
         self.assertEqual(obj1, col_schema.decode(self.col_xml_root, namespaces=self.col_nsmap))
 
         # With lxml.etree namespaces are mapped
-        self.assertEqual(obj1, col_schema.decode(self.col_lxml_root))
+        if self.col_lxml_root is not None:
+            self.assertEqual(obj1, col_schema.decode(self.col_lxml_root))
 
         root = col_schema.encode(obj2, path='./col:collection', namespaces=self.col_nsmap)
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
@@ -416,7 +437,8 @@ class TestConverters(unittest.TestCase):
         self.assertEqual(obj1, col_schema.decode(self.col_xml_root, namespaces=self.col_nsmap))
 
         # With lxml.etree namespaces are mapped
-        self.assertEqual(obj1, col_schema.decode(self.col_lxml_root))
+        if self.col_lxml_root is not None:
+            self.assertEqual(obj1, col_schema.decode(self.col_lxml_root))
 
         root = col_schema.encode(obj2, path='./col:collection', namespaces=self.col_nsmap)
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
@@ -443,7 +465,8 @@ class TestConverters(unittest.TestCase):
         self.assertEqual(obj1, col_schema.decode(self.col_xml_root, namespaces=self.col_nsmap))
 
         # With lxml.etree namespaces are mapped
-        self.assertEqual(obj1, col_schema.decode(self.col_lxml_root))
+        if self.col_lxml_root is not None:
+            self.assertEqual(obj1, col_schema.decode(self.col_lxml_root))
 
         root = col_schema.encode(obj2, path='./col:collection', namespaces=self.col_nsmap)
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
@@ -479,9 +502,10 @@ class TestConverters(unittest.TestCase):
         ))
 
         # With lxml.etree namespaces are mapped
-        self.assertIsNone(etree_elements_assert_equal(
-            obj1, col_schema.decode(self.col_lxml_root)
-        ))
+        if self.col_lxml_root is not None:
+            self.assertIsNone(etree_elements_assert_equal(
+                obj1, col_schema.decode(self.col_lxml_root)
+            ))
 
         root = col_schema.encode(obj2, path='./col:collection', namespaces=self.col_nsmap)
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
@@ -489,6 +513,106 @@ class TestConverters(unittest.TestCase):
         root = col_schema.encode(obj2)  # No namespace unmap is required
         self.assertIsNone(etree_elements_assert_equal(self.col_xml_root, root, strict=False))
 
+    def test_decode_encode_with_default_namespace(self):
+        # Using default namespace and qualified form for elements
+        qualified_col_xsd = self.casepath('examples/collection/collection5.xsd')
+        col_schema = XMLSchema(qualified_col_xsd, converter=BadgerFishConverter)
+
+        default_xml_filename = self.casepath('examples/collection/collection-default.xml')
+        obj1 = col_schema.decode(default_xml_filename)
+        self.assertIn('@xmlns', obj1)
+        self.assertEqual(repr(obj1).count("'@xmlns'"), 1)
+        self.assertEqual(obj1['@xmlns'], {'$': 'http://example.com/ns/collection',
+                                          'xsi': 'http://www.w3.org/2001/XMLSchema-instance'})
+
+        root = col_schema.encode(obj1)
+        default_xml_root = etree_parse(default_xml_filename).getroot()
+        self.assertIsNone(etree_elements_assert_equal(default_xml_root, root, strict=False))
+
+    def test_simple_content__issue_315(self):
+        schema = XMLSchema(self.casepath('issues/issue_315/issue_315_simple.xsd'))
+        converters = (
+            XMLSchemaConverter, XMLSchemaConverter(preserve_root=True),
+            BadgerFishConverter, AbderaConverter, JsonMLConverter,
+            UnorderedConverter, ColumnarConverter, DataElementConverter
+        )
+
+        for k in range(1, 6):
+            xml_filename = self.casepath(f'issues/issue_315/issue_315-{k}.xml')
+            if k < 3:
+                self.assertIsNone(schema.validate(xml_filename), xml_filename)
+            else:
+                self.assertFalse(schema.is_valid(xml_filename), xml_filename)
+
+        for k in (1, 2):
+            xml_filename = self.casepath(f'issues/issue_315/issue_315-{k}.xml')
+            xml_tree = etree_parse(xml_filename).getroot()
+            for converter in converters:
+                obj = schema.decode(xml_filename, converter=converter)
+                root = schema.encode(obj, converter=converter)
+                self.assertIsNone(etree_elements_assert_equal(xml_tree, root))
+
+    def test_mixed_content__issue_315(self):
+        schema = XMLSchema(self.casepath('issues/issue_315/issue_315_mixed.xsd'))
+        losslessly_converters = (JsonMLConverter, DataElementConverter)
+        default_converters = (
+            XMLSchemaConverter(cdata_prefix='#'),
+            UnorderedConverter(cdata_prefix='#'),  # BadgerFishConverter, ColumnarConverter,
+        )
+
+        for k in range(1, 6):
+            xml_filename = self.casepath(f'issues/issue_315/issue_315-{k}.xml')
+            self.assertIsNone(schema.validate(xml_filename), xml_filename)
+
+        for k in range(1, 6):
+            xml_filename = self.casepath(f'issues/issue_315/issue_315-{k}.xml')
+            xml_tree = etree_parse(xml_filename).getroot()
+            for converter in losslessly_converters:
+                obj = schema.decode(xml_filename, converter=converter)
+                root = schema.encode(obj, converter=converter)
+                self.assertIsNone(etree_elements_assert_equal(xml_tree, root, strict=False))
+
+        for k in range(1, 6):
+            xml_filename = self.casepath(f'issues/issue_315/issue_315-{k}.xml')
+            xml_tree = etree_parse(xml_filename).getroot()
+            for converter in default_converters:
+                obj = schema.decode(xml_filename, converter=converter)
+                root = schema.encode(obj, converter=converter, indent=0)
+                if k < 4:
+                    self.assertIsNone(etree_elements_assert_equal(xml_tree, root, strict=False))
+                    continue
+
+                if k == 4:
+                    self.assertEqual(obj, {'@xmlns:tst': 'http://xmlschema.test/ns',
+                                           '@a1': 'foo', 'e2': [None, None], '#1': 'bar'})
+                    self.assertEqual(len(root), 2)
+                else:
+                    self.assertEqual(obj, {'@xmlns:tst': 'http://xmlschema.test/ns',
+                                           '@a1': 'foo', 'e2': [None], '#1': 'bar'})
+                    self.assertEqual(len(root), 1)
+
+                text = etree_tostring(root, namespaces={'tst': 'http://xmlschema.test/ns'})
+                self.assertEqual(len(text.split('bar')), 2)
+
+        for k in range(1, 6):
+            xml_filename = self.casepath(f'issues/issue_315/issue_315-{k}.xml')
+            xml_tree = etree_parse(xml_filename).getroot()
+            obj = schema.decode(xml_filename, converter=BadgerFishConverter)
+            root = schema.encode(obj, converter=BadgerFishConverter, indent=0)
+            if k < 4:
+                self.assertIsNone(etree_elements_assert_equal(xml_tree, root, strict=False))
+                continue
+
+            if k == 4:
+                self.assertEqual(obj, {'@xmlns': {'tst': 'http://xmlschema.test/ns'},
+                                       'tst:e1': {'@a1': 'foo', 'e2': [{}, {}], '$1': 'bar'}})
+            else:
+                self.assertEqual(obj, {'@xmlns': {'tst': 'http://xmlschema.test/ns'},
+                                       'tst:e1': {'@a1': 'foo', 'e2': [{}], '$1': 'bar'}})
+
+            text = etree_tostring(root, namespaces={'tst': 'http://xmlschema.test/ns'})
+            self.assertEqual(len(text.split('bar')), 2)
+
 
 if __name__ == '__main__':
     import platform
