@@ -126,11 +126,11 @@ class TestNamespaceMapper(unittest.TestCase):
 
         with self.assertRaises(ValueError) as ctx:
             mapper.map_qname('{%selement' % XSD_NAMESPACE)
-        self.assertIn("wrong format", str(ctx.exception))
+        self.assertIn("invalid value", str(ctx.exception))
 
         with self.assertRaises(ValueError) as ctx:
             mapper.map_qname('{%s}element}' % XSD_NAMESPACE)
-        self.assertIn("wrong format", str(ctx.exception))
+        self.assertIn("invalid value", str(ctx.exception))
 
         with self.assertRaises(TypeError) as ctx:
             mapper.map_qname(None)
@@ -151,7 +151,7 @@ class TestNamespaceMapper(unittest.TestCase):
 
         with self.assertRaises(ValueError) as ctx:
             mapper.unmap_qname('xs::element')
-        self.assertIn("wrong format", str(ctx.exception))
+        self.assertIn("invalid value", str(ctx.exception))
 
         with self.assertRaises(TypeError) as ctx:
             mapper.unmap_qname(None)
@@ -166,6 +166,9 @@ class TestNamespaceMapper(unittest.TestCase):
         self.assertEqual(mapper.unmap_qname('element'), '{foo}element')
         self.assertEqual(mapper.unmap_qname('element', name_table=['element']), 'element')
 
+        mapper.strip_namespaces = True
+        self.assertEqual(mapper.unmap_qname('element'), 'element')
+
     def test_transfer(self):
         mapper = NamespaceMapper(namespaces={'xs': XSD_NAMESPACE, 'xsi': XSI_NAMESPACE})
 
@@ -205,6 +208,17 @@ class TestNamespaceView(unittest.TestCase):
         ns_view = NamespaceView(qnames, 'tns0')
         self.assertEqual(repr(ns_view), "NamespaceView({'name0': 0})")
 
+    def test_contains(self):
+        qnames = {'{tns0}name0': 0, '{tns1}name1': 1, 'name2': 2}
+        ns_view = NamespaceView(qnames, 'tns1')
+
+        self.assertIn('name1', ns_view)
+        self.assertNotIn('{tns1}name1', ns_view)
+        self.assertNotIn('{tns0}name0', ns_view)
+        self.assertNotIn('name0', ns_view)
+        self.assertNotIn('name2', ns_view)
+        self.assertNotIn(1, ns_view)
+
     def test_as_dict(self):
         qnames = {'{tns0}name0': 0, '{tns1}name1': 1, '{tns1}name2': 2, 'name3': 3}
         ns_view = NamespaceView(qnames, 'tns1')
