@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+# !/usr/bin/env python
 #
 # Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
@@ -10,10 +10,13 @@
 #
 import unittest
 import os
+import base64
 import json
+import math
 from decimal import Decimal
 from collections.abc import MutableMapping, MutableSequence, Set
-import base64
+from textwrap import dedent
+from xml.etree import ElementTree
 
 try:
     import lxml.etree as lxml_etree
@@ -23,12 +26,11 @@ except ImportError:
 from elementpath import datatypes
 import xmlschema
 from xmlschema import XMLSchemaValidationError, ParkerConverter, BadgerFishConverter, \
-    AbderaConverter, JsonMLConverter, ColumnarConverter
+    AbderaConverter, JsonMLConverter, ColumnarConverter, ElementData
 
-from xmlschema.names import XSD_STRING
-from xmlschema.etree import ElementTree
+from xmlschema.names import XSD_STRING, XSI_NIL
 from xmlschema.converters import UnorderedConverter
-from xmlschema.validators import XMLSchema11
+from xmlschema.validators import XMLSchema11, ModelVisitor
 from xmlschema.testing import XsdValidatorTestCase, etree_elements_assert_equal
 
 VEHICLES_DICT = {
@@ -326,6 +328,7 @@ DATA_DICT = {
     ],
     'simple_boolean': [True, False],
     'date_and_time': '2020-03-05T23:04:10.047',  # xs:dateTime is not decoded for default
+    'list_of_floats': [float('inf'), float('-inf')],
 }
 
 
@@ -372,7 +375,9 @@ class TestDecoding(XsdValidatorTestCase):
         xml_dict = self.vh_schema.to_dict(vh_xml_tree, namespaces=self.vh_namespaces)
         self.assertEqual(xml_dict, VEHICLES_DICT)
 
-        xml_dict = xmlschema.to_dict(vh_xml_tree, self.vh_schema.url, namespaces=self.vh_namespaces)
+        xml_dict = xmlschema.to_dict(
+            vh_xml_tree, self.vh_schema.url, namespaces=self.vh_namespaces
+        )
         self.assertEqual(xml_dict, VEHICLES_DICT)
 
         xml_dict = self.col_schema.to_dict(col_xml_tree)
@@ -408,7 +413,7 @@ class TestDecoding(XsdValidatorTestCase):
 
     def test_date_decoding(self):
         # Issue #136
-        schema = xmlschema.XMLSchema("""<?xml version="1.0" encoding="utf-8"?>
+        schema = self.schema_class("""<?xml version="1.0" encoding="utf-8"?>
             <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" version="1.0">
                 <xs:element name="Date">
                     <xs:simpleType>
@@ -475,7 +480,7 @@ class TestDecoding(XsdValidatorTestCase):
         with self.assertRaises(XMLSchemaValidationError) as ctx:
             schema.decode('<root>ns0:foo</root>')
         self.assertIn("failed validating 'ns0:foo'", str(ctx.exception))
-        self.assertIn("Reason: unmapped prefix 'ns0' on QName", str(ctx.exception))
+        self.assertIn("Reason: unmapped prefix 'ns0' in a QName", str(ctx.exception))
         self.assertIn("Path: /root", str(ctx.exception))
 
         xml_data = '<root name="ns0:bar" xmlns:ns0="http://xmlschema.test/0">ns0:foo</root>'
@@ -489,7 +494,7 @@ class TestDecoding(XsdValidatorTestCase):
         with self.assertRaises(XMLSchemaValidationError) as ctx:
             schema.decode('<root name="ns0:bar">foo</root>')
         self.assertIn("failed validating 'ns0:bar'", str(ctx.exception))
-        self.assertIn("unmapped prefix 'ns0' on QName", str(ctx.exception))
+        self.assertIn("unmapped prefix 'ns0' in a QName", str(ctx.exception))
         self.assertIn("Path: /root", str(ctx.exception))
 
     def test_json_dump_and_load(self):
@@ -627,9 +632,9 @@ class TestDecoding(XsdValidatorTestCase):
 
     def test_non_global_schema_path(self):
         # Issue #157
-        xs = xmlschema.XMLSchema("""<?xml version="1.0" encoding="UTF-8"?>
-        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" 
-                xmlns:foo="http://example.com/foo" 
+        xs = self.schema_class("""<?xml version="1.0" encoding="UTF-8"?>
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
+                xmlns:foo="http://example.com/foo"
                 targetNamespace="http://example.com/foo">
             <xs:complexType name="type1">
                 <xs:sequence>
@@ -756,7 +761,7 @@ class TestDecoding(XsdValidatorTestCase):
                 self.col_xml_file, converter=ColumnarConverter, attr_prefix='-',
             )
         self.assertEqual(str(ctx.exception),
-                         "attr_prefix can be the empty string or a single/double underscore")
+                         "'attr_prefix' can be the empty string or a single/double underscore")
 
     def test_dict_granularity(self):
         """Based on Issue #22, test to make sure an xsd indicating list with
@@ -764,22 +769,66 @@ class TestDecoding(XsdValidatorTestCase):
         xsd_string = self.casepath('issues/issue_022/xsd_string.xsd')
         xml_string_1 = self.casepath('issues/issue_022/xml_string_1.xml')
         xml_string_2 = self.casepath('issues/issue_022/xml_string_2.xml')
-        xsd_schema = xmlschema.XMLSchema(xsd_string)
+        xsd_schema = self.schema_class(xsd_string)
         xml_data_1 = xsd_schema.to_dict(xml_string_1)
         xml_data_2 = xsd_schema.to_dict(xml_string_2)
         self.assertTrue(isinstance(xml_data_1['bar'], type(xml_data_2['bar'])),
                         msg="XSD with an array that return a single element from "
                             "xml must still yield a list.")
 
-    def test_any_type(self):
-        any_type = xmlschema.XMLSchema.meta_schema.types['anyType']
+    def test_any_type_decoding(self):
+        any_type = self.schema_class.meta_schema.types['anyType']
         xml_data_1 = ElementTree.Element('dummy')
         self.assertIsNone(any_type.decode(xml_data_1))
         xml_data_2 = ElementTree.fromstring('<root>\n    <child_1/>\n    <child_2/>\n</root>')
-        self.assertIsNone(any_type.decode(xml_data_2))  # Currently no decoding yet
+
+        # Fix for processContents='lax' (issue 273, previously result was None)
+        self.assertEqual(any_type.decode(xml_data_2), {'child_1': [None], 'child_2': [None]})
+
+    def test_skip_wildcard_decoding(self):
+        schema = self.schema_class("""<?xml version="1.0" encoding="UTF-8"?>
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+                <xs:element name="foo">
+                    <xs:complexType>
+                        <xs:sequence>
+                            <xs:any processContents="skip"/>
+                        </xs:sequence>
+                    </xs:complexType>
+                </xs:element>
+            </xs:schema>""")
+
+        self.assertIsNone(schema.decode('<foo><bar/></foo>'))
+
+        obj = schema.decode('<foo><bar/></foo>', process_skipped=True)
+        self.assertEqual(obj, {'bar': None})
+
+        root = schema.encode(obj)
+        self.assertEqual(root.tag, 'foo')
+        self.assertEqual(len(root), 0)
+
+        root = schema.encode(obj, process_skipped=True)
+        self.assertEqual(root.tag, 'foo')
+        self.assertEqual(len(root), 1)
+        self.assertEqual(root[0].tag, 'bar')
+
+    def test_any_type_decoding__issue_273(self):
+        schema = self.schema_class(self.casepath('issues/issue_273/issue_273.xsd'))
+        data = schema.to_dict(self.casepath('issues/issue_273/issue_273.xml'))
+
+        self.assertDictEqual(data, {
+            '@x': 'QA01_sequence',
+            '@y': 2500,
+            'QA01': [
+                {'qa01_elem01': ['13'],
+                 'qa01_elem02': ['5139'],
+                 'qa01_elem03': ['170'],
+                 'qa01_elem04': [{'@these': 'attributes', '@get': 'dropped', '$': '0'}],
+                 'qa01_elem05': ['56'],
+                 'qa01_elem06': ['11141178']
+                 }]})
 
     def test_choice_model_decoding__issue_041(self):
-        schema = xmlschema.XMLSchema(self.casepath('issues/issue_041/issue_041.xsd'))
+        schema = self.schema_class(self.casepath('issues/issue_041/issue_041.xsd'))
         data = schema.to_dict(self.casepath('issues/issue_041/issue_041.xml'))
         self.assertEqual(data, {
             '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
@@ -789,7 +838,7 @@ class TestDecoding(XsdValidatorTestCase):
         })
 
     def test_cdata_decoding(self):
-        schema = xmlschema.XMLSchema(self.casepath('issues/issue_046/issue_046.xsd'))
+        schema = self.schema_class(self.casepath('issues/issue_046/issue_046.xsd'))
         xml_file = self.casepath('issues/issue_046/issue_046.xml')
         self.assertEqual(
             schema.decode(xml_file, cdata_prefix='#'),
@@ -819,8 +868,8 @@ class TestDecoding(XsdValidatorTestCase):
         self.assertIsInstance(obj, str)
 
         obj = xs.decode('<hex> 9AFD </hex>', binary_types=True)
-        self.assertEqual(obj, '9AFD')
         self.assertIsInstance(obj, datatypes.HexBinary)
+        self.assertEqual(obj.value, b'9AFD')
 
         xs = self.get_schema('<xs:attribute name="hex" type="xs:hexBinary"/>')
 
@@ -829,8 +878,8 @@ class TestDecoding(XsdValidatorTestCase):
         self.assertIsInstance(obj, str)
 
         obj = xs.attributes['hex'].decode(' 9AFD ', binary_types=True)
-        self.assertEqual(obj, '9AFD')
         self.assertIsInstance(obj, datatypes.HexBinary)
+        self.assertEqual(obj.value, b'9AFD')
 
     def test_base64_binary_type(self):
         base64_code_type = self.st_schema.types['base64Code']
@@ -844,7 +893,7 @@ class TestDecoding(XsdValidatorTestCase):
         xs = self.get_schema('<xs:attribute name="b64" type="xs:base64Binary"/>')
 
         obj = xs.attributes['b64'].decode(base64_value.decode())
-        self.assertEqual(obj, expected_value)
+        self.assertEqual(obj, str(expected_value))
         self.assertIsInstance(obj, str)
 
         obj = xs.attributes['b64'].decode(base64_value.decode(), binary_types=True)
@@ -855,7 +904,7 @@ class TestDecoding(XsdValidatorTestCase):
         xs = self.get_schema('<xs:element name="b64" type="xs:base64Binary"/>')
 
         obj = xs.decode('<b64>{}</b64>'.format(base64_value.decode()))
-        self.assertEqual(obj, expected_value)
+        self.assertEqual(obj, str(expected_value))
         self.assertIsInstance(obj, str)
 
         obj = xs.decode('<b64>{}</b64>'.format(base64_value.decode()), binary_types=True)
@@ -918,7 +967,7 @@ class TestDecoding(XsdValidatorTestCase):
             </xs:complexType>
         </xs:schema>
         """
-        xsd_schema = xmlschema.XMLSchema(xsd_string)
+        xsd_schema = self.schema_class(xsd_string)
         xml_string_1 = "<foo><bar>0</bar></foo>"
         xml_string_2 = """<?xml version="1.0" encoding="UTF-8"?>
         <foo xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
@@ -930,9 +979,85 @@ class TestDecoding(XsdValidatorTestCase):
         obj = xsd_schema.decode(xml_string_2, use_defaults=False)
         self.check_etree_elements(ElementTree.fromstring(xml_string_2), xsd_schema.encode(obj))
 
+    def test_keep_empty(self):
+        schema = self.schema_class(self.casepath('issues/issue_322/issue_322.xsd'))
+        xml_file = self.casepath('issues/issue_322/issue_322.xml')
+
+        data = schema.decode(xml_file)
+        self.assertEqual(data, {
+            '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
+            'emptystring': None,
+            'nillstring': {'@xsi:nil': 'true'}
+        })
+
+        data = schema.decode(xml_file, keep_empty=True)
+        self.assertEqual(data, {
+            '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
+            'emptystring': '',
+            'nillstring': {'@xsi:nil': 'true'}
+        })
+
+        schema = self.schema_class(dedent("""\
+        <?xml version="1.0"?>
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+          <xs:element name="note">
+            <xs:complexType>
+              <xs:sequence>
+                <xs:element name="emptiable" type="xs:string"/>
+                <xs:element name="filled">
+                  <xs:simpleType>
+                    <xs:restriction base="xs:string">
+                      <xs:minLength value="1"/>
+                    </xs:restriction>
+                  </xs:simpleType>
+                </xs:element>
+                <xs:element name="number" type="xs:int"/>
+              </xs:sequence>
+            </xs:complexType>
+          </xs:element>
+        </xs:schema>"""))
+
+        xml_data = "<note><emptiable/><filled/><number/></note>"
+        data, errors = schema.decode(xml_data, validation='lax', keep_empty=True)
+        self.assertEqual(data, {'emptiable': '', 'filled': '', 'number': None})
+        self.assertEqual(len(errors), 2)
+
+        data = schema.decode(xml_data, validation='skip', keep_empty=True)
+        self.assertEqual(data, {'emptiable': '', 'filled': '', 'number': ''})
+
+    def test_element_hook__issue_322(self):
+        schema = self.schema_class(self.casepath('issues/issue_322/issue_322.xsd'))
+        xml_file = self.casepath('issues/issue_322/issue_322.xml')
+
+        def element_hook(element_data: ElementData, *_args):
+            if not element_data.attributes:
+                return element_data
+
+            return ElementData(
+                element_data.tag,
+                element_data.text,
+                element_data.content,
+                [x for x in element_data.attributes if x[0] != XSI_NIL]
+            )
+
+        data = schema.decode(xml_file, element_hook=element_hook)
+        self.assertEqual(data, {
+            '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
+            'emptystring': None,
+            'nillstring': None,
+        })
+
+        # Resolution for issue 322
+        data = schema.decode(xml_file, keep_empty=True, element_hook=element_hook)
+        self.assertEqual(data, {
+            '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
+            'emptystring': '',
+            'nillstring': None,
+        })
+
     def test_default_namespace__issue_077(self):
-        xs = xmlschema.XMLSchema("""<?xml version="1.0" encoding="UTF-8"?>
-        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" 
+        xs = self.schema_class("""<?xml version="1.0" encoding="UTF-8"?>
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
             targetNamespace="http://example.com/foo">
           <xs:element name="foo" type="xs:string" />
         </xs:schema>""")
@@ -1090,13 +1215,13 @@ class TestDecoding(XsdValidatorTestCase):
             </xs:schema>"""
 
         schema = self.schema_class(xsd_text)
-        result = schema.to_dict(
-            """<root xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
+        xml_source = """<root xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
               <ExtensionSub xsi:type="ExtensionType">
                 <Content>my content</Content>
               </ExtensionSub>
             </root>"""
-        )
+
+        result = schema.to_dict(xml_source)
         expected = {
             "@xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
             "ExtensionSub": {
@@ -1106,6 +1231,11 @@ class TestDecoding(XsdValidatorTestCase):
         }
         self.assertEqual(result, expected)
 
+        root = schema.to_etree(result)
+        self.assertIsNone(etree_elements_assert_equal(
+            root, ElementTree.XML(xml_source), strict=False
+        ))
+
     def test_issue_190(self):
         # Changed is_single() for XsdElement to check also parent group.
         schema = self.schema_class(self.casepath('issues/issue_190/issue_190.xsd'))
@@ -1152,6 +1282,187 @@ class TestDecoding(XsdValidatorTestCase):
         self.assertIsInstance(obj, str)
         self.assertEqual(json_data.encode("utf-8"), b'"\\u00f8\\u00e6\\u00e5"')
 
+    def test_float_decoding(self):
+        for float_type in [self.schema_class.meta_schema.types['float'],
+                           self.schema_class.meta_schema.types['double']]:
+            self.assertAlmostEqual(float_type.decode('-1'), -1.0)
+            self.assertAlmostEqual(float_type.decode('19.7'), 19.7)
+            self.assertAlmostEqual(float_type.decode('INF'), float('inf'))
+            self.assertAlmostEqual(float_type.decode('-INF'), float('-inf'))
+            self.assertNotAlmostEqual(float_type.decode('INF'), float('-inf'))
+            self.assertTrue(math.isnan(float_type.decode('NaN')))
+
+    def test_no_process_namespaces__issue_314(self):
+        xsd_file = self.casepath('issues/issue_314/issue_314.xsd')
+        xml_file = self.casepath('issues/issue_314/issue_314.xml')
+        schema = self.schema_class(xsd_file)
+
+        result = schema.decode(xml_file)
+        self.assertEqual(
+            result,
+            {'@xmlns:p': 'my_namespace',
+             '@xmlns:b': 'http://www.w3.org/2001/XMLSchema-instance',
+             'p:container': {
+                 'p:item': [{
+                     '@b:type': 'p:ConcreteContainterItemInfo',
+                     '@attr_2': 'value_2'}
+                 ]}}
+        )
+
+        result = schema.decode(xml_file, process_namespaces=False)
+        self.assertEqual(
+            result,
+            {'{my_namespace}container': {
+                '{my_namespace}item': [{
+                    '@{http://www.w3.org/2001/XMLSchema-instance}type':
+                        'p:ConcreteContainterItemInfo',
+                    '@attr_2': 'value_2'}]}}
+        )
+
+    def test_mixed_content_decode__issue_334(self):
+        schema = self.schema_class("""
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+
+            <xs:complexType name="mixedContentType" mixed="true">
+                <xs:sequence>
+                    <xs:element name="elem1" maxOccurs="unbounded"/>
+                </xs:sequence>
+            </xs:complexType>
+
+            <xs:element name="foo">
+              <xs:complexType>
+                <xs:complexContent>
+                  <xs:extension base="mixedContentType">
+                    <xs:attribute name="bar" type="xs:string" />
+                  </xs:extension>
+                </xs:complexContent>
+              </xs:complexType>
+            </xs:element>
+
+            <xs:element name="root" type="mixedContentType" />
+
+        </xs:schema>
+        """)
+
+        result = schema.decode('<root>text1<elem1/>tail1<elem1/>tail2</root>')
+        self.assertEqual(result, {'elem1': [None, None]})
+
+        result = schema.decode('<root>text1<elem1/>tail1<elem1/>tail2</root>', cdata_prefix='#')
+        self.assertEqual(result, {'#1': 'text1',
+                                  'elem1': [None, None],
+                                  '#2': 'tail1',
+                                  '#3': 'tail2'})
+        result = schema.decode('<root>text1<elem1/>tail1<elem1/>tail2</root>', cdata_prefix='')
+        self.assertEqual(result, {'1': 'text1',
+                                  'elem1': [None, None],
+                                  '2': 'tail1',
+                                  '3': 'tail2'})
+
+        xsd_file = self.casepath('issues/issue_334/issue_334.xsd')
+        xml_file = self.casepath('issues/issue_334/issue_334.xml')
+        xs = self.schema_class(xsd_file)
+        result = xs.decode(xml_file)
+        body_text = result['Demonstrative_Example'][0]['Body_Text']
+
+        expected = ['The snippet of code below establishes a new cookie to hold the sessionID.',
+                    'The HttpOnly flag is not set for the cookie. An attacker who can perform '
+                    'XSS could insert malicious script such as:',
+                    'When the client loads and executes this script, it makes a request to the '
+                    'attacker-controlled web site. The attacker can then log the request and '
+                    'steal the cookie.',
+                    'To mitigate the risk, use the setHttpOnly(true) method.']
+        self.assertListEqual(body_text, expected)
+
+        result = xs.decode(xml_file, preserve_root=True)
+        body_text = result['Demonstrative_Examples']['Demonstrative_Example'][0]['Body_Text']
+        self.assertListEqual(body_text, expected)
+
+    def test_fill_missing_elements__issue_341(self):
+        xsd_file = self.casepath('issues/issue_341/issue_341.xsd')
+        xml_file = self.casepath('issues/issue_341/issue_341.xml')
+        schema = self.schema_class(xsd_file)
+
+        expected = {'TEST_EL': [
+            {'@Date': '2022-10-03',
+             'TEST_EL_2': {
+                 'exists_in_xml': {
+                     '@test_attr': 'test_value_attr', '@test_attr_2': None
+                 }
+             }}
+        ]}
+        xml_dict = schema.decode(xml_file, fill_missing=True)
+        self.assertDictEqual(xml_dict, expected)
+
+        def fill_missing_content(element_data: ElementData, _xsd_element, xsd_type):
+            group = xsd_type.model_group
+            if group is None:
+                return element_data  # an element with simple content
+
+            filled_content = []
+            model = ModelVisitor(group)
+            xsd_element = None
+            for name, value, xsd_element in element_data.content:
+                if isinstance(name, int) or xsd_element is None:
+                    filled_content.append((name, value, xsd_element))
+                    continue
+
+                while model.element is not None:
+                    if model.element is xsd_element:
+                        filled_content.append((name, value, xsd_element))
+                        for _err in model.advance(True):
+                            pass
+                        break
+
+                    if model.element.max_occurs != 0:
+                        filled_content.append((model.element.name, None, model.element))
+                    for _err in model.advance(False):
+                        pass
+                else:
+                    filled_content.append((name, value, xsd_element))
+
+            while model.element is not None:
+                if model.element is not xsd_element and model.element.max_occurs != 0:
+                    filled_content.append((model.element.name, None, model.element))
+                for _err in model.advance(False):
+                    pass
+
+            return ElementData(
+                element_data.tag,
+                element_data.text,
+                filled_content,
+                element_data.attributes
+            )
+
+        expected = {'TEST_EL': [
+            {'@Date': '2022-10-03',
+             'TEST_EL_2': {
+                 'exists_in_xml': {
+                     '@test_attr': 'test_value_attr', '@test_attr_2': None
+                 },
+                 'not_exists_in_xml': None
+             }}
+        ]}
+        xml_dict = schema.decode(xml_file, element_hook=fill_missing_content, fill_missing=True)
+        self.assertDictEqual(xml_dict, expected)
+
+        # Resolving more complex schemas requires more checks in hook function
+        xsd_file = self.casepath('issues/issue_341/issue_341-ext.xsd')
+        schema = self.schema_class(xsd_file)
+
+        expected = {'TEST_EL': [
+            {'@Date': '2022-10-03',
+             'TEST_EL_2': {
+                 'exists_in_xml': {
+                     '@test_attr': 'test_value_attr', '@test_attr_2': None
+                 },
+                 'not_exists_in_xml': None,
+                 'choice_elem1': None,
+                 'choice_elem2': None,  # this is wrong (at most one element for a choice)
+             }}
+        ]}
+        xml_dict = schema.decode(xml_file, element_hook=fill_missing_content, fill_missing=True)
+        self.assertDictEqual(xml_dict, expected)
+
 
 class TestDecoding11(TestDecoding):
     schema_class = XMLSchema11
