@@ -12,6 +12,8 @@ import unittest
 import os
 import sys
 import decimal
+from textwrap import dedent
+from xml.etree import ElementTree
 
 try:
     import lxml.etree as lxml_etree
@@ -21,7 +23,6 @@ except ImportError:
 import xmlschema
 from xmlschema import XMLSchemaValidationError
 
-from xmlschema.etree import ElementTree
 from xmlschema.validators import XMLSchema11
 from xmlschema.testing import XsdValidatorTestCase
 
@@ -112,24 +113,97 @@ class TestValidation(XsdValidatorTestCase):
 
     def test_max_depth_argument(self):
         schema = self.schema_class(self.col_xsd_file)
-        self.assertEqual(
-            schema.decode(self.col_xml_file, max_depth=1),
-            {'@xmlns:col': 'http://example.com/ns/collection',
-             '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
-             '@xsi:schemaLocation': 'http://example.com/ns/collection collection.xsd'})
+        invalid_col_xml_file = self.casepath('examples/collection/collection-1_error.xml')
+        self.assertFalse(schema.is_valid(invalid_col_xml_file))
+        self.assertTrue(schema.is_valid(invalid_col_xml_file, max_depth=1))
+        self.assertTrue(schema.is_valid(invalid_col_xml_file, max_depth=2))
+        self.assertFalse(schema.is_valid(invalid_col_xml_file, max_depth=3))
 
-        xmlschema.limits.MAX_XML_DEPTH = 1
-        with self.assertRaises(XMLSchemaValidationError):
-            schema.decode(self.col_xml_file)
-        xmlschema.limits.MAX_XML_DEPTH = 9999
+        root = ElementTree.parse(invalid_col_xml_file).getroot()
+        xsd_element = schema.elements['collection']
+        self.assertFalse(xsd_element.is_valid(root))
+        self.assertTrue(xsd_element.is_valid(root, max_depth=1))
+        self.assertTrue(xsd_element.is_valid(root, max_depth=2))
+        self.assertFalse(xsd_element.is_valid(root, max_depth=3))
 
-        self.assertEqual(
-            schema.decode(self.col_xml_file, max_depth=2),
-            {'@xmlns:col': 'http://example.com/ns/collection',
-             '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
-             '@xsi:schemaLocation': 'http://example.com/ns/collection collection.xsd',
-             'object': [{'@id': 'b0836217462', '@available': True},
-                        {'@id': 'b0836217463', '@available': True}]})
+        # Need to provide namespace explicitly because the default namespace
+        # is set with xpathDefaultNamespace, that is '' in this case.
+        xsd_element = schema.find('collection/object', namespaces={'': schema.target_namespace})
+
+        self.assertTrue(xsd_element.is_valid(root[0]))
+        self.assertFalse(xsd_element.is_valid(root[1]))
+        self.assertTrue(xsd_element.is_valid(root[1], max_depth=1))
+        self.assertFalse(xsd_element.is_valid(root[1], max_depth=2))
+
+    def test_extra_validator_argument(self):
+        # Related to issue 227
+
+        def bikes_validator(elem, xsd_element_):
+            if elem.tag == '{http://example.com/vehicles}bike' and \
+                    xsd_element_.name == elem.tag and \
+                    elem.attrib['make'] != 'Harley-Davidson':
+                raise XMLSchemaValidationError(xsd_element_, elem, 'not an Harley-Davidson')
+
+        with self.assertRaises(XMLSchemaValidationError) as ec:
+            self.vh_schema.validate(self.vh_xml_file, extra_validator=bikes_validator)
+        self.assertIn('Reason: not an Harley-Davidson', str(ec.exception))
+
+        root = ElementTree.parse(self.vh_xml_file).getroot()
+        xsd_element = self.vh_schema.elements['vehicles']
+        with self.assertRaises(XMLSchemaValidationError) as ec:
+            xsd_element.validate(root, extra_validator=bikes_validator)
+        self.assertIn('Reason: not an Harley-Davidson', str(ec.exception))
+
+        with self.assertRaises(XMLSchemaValidationError) as ec:
+            self.vh_schema.decode(self.vh_xml_file, extra_validator=bikes_validator)
+        self.assertIn('Reason: not an Harley-Davidson', str(ec.exception))
+
+        def bikes_validator(elem, xsd_element_):
+            if elem.tag == '{http://example.com/vehicles}bike' and \
+                    xsd_element_.name == elem.tag and \
+                    elem.attrib['make'] != 'Harley-Davidson':
+                yield XMLSchemaValidationError(xsd_element_, elem, 'not an Harley-Davidson')
+
+        with self.assertRaises(XMLSchemaValidationError) as ec:
+            self.vh_schema.validate(self.vh_xml_file, extra_validator=bikes_validator)
+        self.assertIn('Reason: not an Harley-Davidson', str(ec.exception))
+
+    def test_path_argument(self):
+        schema = xmlschema.XMLSchema(self.casepath('examples/vehicles/vehicles.xsd'))
+
+        self.assertTrue(schema.is_valid(self.vh_xml_file, path='*'))
+        self.assertTrue(schema.is_valid(self.vh_xml_file, path='/vh:vehicles'))
+        self.assertTrue(schema.is_valid(self.vh_xml_file, path='/vh:vehicles/vh:cars'))
+        self.assertTrue(schema.is_valid(self.vh_xml_file, path='vh:cars'))
+        self.assertTrue(schema.is_valid(self.vh_xml_file, path='/vh:vehicles/vh:cars/vh:car'))
+        self.assertTrue(schema.is_valid(self.vh_xml_file, path='.//vh:car'))
+
+        self.assertTrue(schema.is_valid(self.vh_xml_file, path='xs:vehicles'))
+        self.assertFalse(
+            schema.is_valid(self.vh_xml_file, path='xs:vehicles', allow_empty=False)
+        )
+
+    def test_schema_path_argument__issue_326(self):
+        schema = xmlschema.XMLSchema(self.casepath('examples/vehicles/vehicles.xsd'))
+        document = ElementTree.parse(self.vh_xml_file)
+
+        entries = document.findall('vh:cars', {'vh': 'http://example.com/vehicles'})
+        self.assertListEqual(entries, [document.getroot()[0]])
+        for entry in entries:
+            self.assertTrue(schema.is_valid(
+                entry,
+                schema_path='/vh:vehicles/vh:cars',
+                namespaces={'vh': 'http://example.com/vehicles'}
+            ))
+
+        entries = document.findall('vh:cars/vh:car', {'vh': 'http://example.com/vehicles'})
+        self.assertListEqual(entries, list(document.getroot()[0][:]))
+        for entry in entries:
+            self.assertTrue(schema.is_valid(
+                entry,
+                schema_path='.//vh:cars/vh:car',
+                namespaces={'vh': 'http://example.com/vehicles'}
+            ))
 
     def test_issue_064(self):
         self.check_validity(self.st_schema, '<name xmlns="ns"></name>', False)
@@ -198,7 +272,7 @@ class TestValidation(XsdValidatorTestCase):
                 <xs:element name="elem2" type="xs:string"/>
                 <xs:element name="elem3" type="xs:string"/>
                 <xs:element name="elem4" type="xs:string"/>
-                
+
                 <xs:element name="root" type="enumType"/>
 
                 <xs:simpleType name="enumType">
@@ -207,7 +281,7 @@ class TestValidation(XsdValidatorTestCase):
                         <xs:enumeration value="elem2"/>
                         <xs:enumeration value="tns1:other1"/>
                         <xs:enumeration value="elem3"/>
-                        <xs:enumeration value="tns2:other2"/>                        
+                        <xs:enumeration value="tns2:other2"/>
                         <xs:enumeration value="elem4"/>
                     </xs:restriction>
                 </xs:simpleType>
@@ -217,11 +291,10 @@ class TestValidation(XsdValidatorTestCase):
         self.check_validity(schema, xml_data, True)
 
     def test_issue_213(self):
-        schema = xmlschema.XMLSchema("""
+        schema = xmlschema.XMLSchema(dedent("""\
         <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
-        <xs:element name="amount" type="xs:decimal"/>
-        </xs:schema>
-        """)
+          <xs:element name="amount" type="xs:decimal"/>
+        </xs:schema>"""))
 
         xml1 = """<?xml version="1.0" encoding="UTF-8"?><amount>0.000000</amount>"""
         self.assertIsInstance(schema.decode(xml1), decimal.Decimal)
@@ -229,6 +302,32 @@ class TestValidation(XsdValidatorTestCase):
         xml2 = """<?xml version="1.0" encoding="UTF-8"?><amount>0.0000000</amount>"""
         self.assertIsInstance(schema.decode(xml2), decimal.Decimal)
 
+    def test_issue_224__validate_malformed_file(self):
+        schema = xmlschema.XMLSchema(dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+              <xs:element name="root" type="xs:string"/>
+            </xs:schema>"""))
+
+        malformed_xml_file = self.casepath('resources/malformed.xml')
+
+        with self.assertRaises(ElementTree.ParseError):
+            schema.is_valid(malformed_xml_file)
+
+    def test_issue_238__validate_bytes_strings(self):
+        schema = xmlschema.XMLSchema(dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+              <xs:element name="value" type="xs:int"/>
+            </xs:schema>"""))
+
+        self.assertTrue(schema.is_valid('<value>11</value>'))
+        self.assertTrue(schema.is_valid(b'<value>\n11\n</value>'))
+
+        with open(self.col_xml_file, 'rb') as fp:
+            col_xml_data = fp.read()
+
+        self.assertIsInstance(col_xml_data, bytes)
+        self.assertTrue(self.col_schema.is_valid(col_xml_data))
+
 
 class TestValidation11(TestValidation):
     schema_class = XMLSchema11
