@@ -9,9 +9,10 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 import unittest
+from textwrap import dedent
+from xml.etree.ElementTree import Element
 
 from xmlschema import XMLSchemaParseError, XMLSchemaModelError
-from xmlschema.etree import etree_element
 from xmlschema.validators import XMLSchema11
 from xmlschema.testing import XsdValidatorTestCase
 
@@ -407,27 +408,14 @@ class TestXsdComplexType(XsdValidatorTestCase):
         self.assertTrue(xsd_type.has_mixed_content())
         self.assertEqual(xsd_type.content_type_label, 'mixed')
 
-    def test_content_type_property(self):
-        schema = self.schema_class("""
-            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
-                <xs:complexType name="type1">
-                    <xs:sequence>
-                        <xs:element name="elem1"/>
-                    </xs:sequence>
-                </xs:complexType>
-            </xs:schema>""")
-
-        xsd_type = schema.types['type1']
-        self.assertIs(xsd_type.content_type, xsd_type.content)
-
     def test_is_empty(self):
         schema = self.check_schema("""
             <xs:complexType name="emptyType1"/>
-            
+
             <xs:complexType name="emptyType2">
                 <xs:sequence/>
             </xs:complexType>
-            
+
             <xs:complexType name="emptyType3">
                 <xs:complexContent>
                     <xs:restriction base="xs:anyType"/>
@@ -439,7 +427,7 @@ class TestXsdComplexType(XsdValidatorTestCase):
                     <xs:element name="elem1"/>
                 </xs:sequence>
             </xs:complexType>
-            
+
             <xs:complexType name="notEmptyType2">
                 <xs:complexContent>
                     <xs:extension base="xs:anyType"/>
@@ -453,6 +441,312 @@ class TestXsdComplexType(XsdValidatorTestCase):
         self.assertFalse(schema.types['notEmptyType1'].is_empty())
         self.assertFalse(schema.types['notEmptyType2'].is_empty())
 
+    def test_restriction_with_empty_particle__issue_323(self):
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+          <xs:complexType name="ED" mixed="true">
+            <xs:complexContent>
+              <xs:restriction base="xs:anyType">
+                <xs:sequence>
+                  <xs:element name="reference" type="xs:string" minOccurs="0" maxOccurs="1"/>
+                  <xs:element name="thumbnail" type="thumbnail" minOccurs="0" maxOccurs="1"/>
+                </xs:sequence>
+              </xs:restriction>
+            </xs:complexContent>
+          </xs:complexType>
+
+          <xs:complexType name="thumbnail" mixed="true">
+            <xs:complexContent>
+              <xs:restriction base="ED">
+                <xs:sequence>
+                    <xs:element name="reference" type="xs:string" minOccurs="0" maxOccurs="0"/>
+                    <xs:element name="thumbnail" type="thumbnail" minOccurs="0" maxOccurs="0"/>
+                </xs:sequence>
+              </xs:restriction>
+            </xs:complexContent>
+          </xs:complexType>
+
+          <xs:complexType name="ST" mixed="true">
+            <xs:complexContent>
+              <xs:restriction base="ED">
+                <xs:sequence>
+                    <xs:element name="reference" type="xs:string" minOccurs="0" maxOccurs="0"/>
+                    <xs:element name="thumbnail" type="ED" minOccurs="0" maxOccurs="0"/>
+                </xs:sequence>
+              </xs:restriction>
+            </xs:complexContent>
+          </xs:complexType>
+        </xs:schema>"""), build=False)
+
+        self.assertIsNone(schema.build())
+        self.assertTrue(schema.built)
+
+    def test_mixed_content_extension__issue_334(self):
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+
+            <xs:complexType name="mixedContentType" mixed="true">
+                <xs:sequence>
+                    <xs:element name="elem1"/>
+                </xs:sequence>
+            </xs:complexType>
+
+            <xs:element name="foo">
+              <xs:complexType>
+                <xs:complexContent>
+                  <xs:extension base="mixedContentType">
+                    <xs:attribute name="bar" type="xs:string" use="required" />
+                  </xs:extension>
+                </xs:complexContent>
+              </xs:complexType>
+            </xs:element>
+
+        </xs:schema>
+        """))
+
+        self.assertTrue(schema.types['mixedContentType'].mixed)
+        self.assertTrue(schema.elements['foo'].type.mixed)
+        self.assertTrue(schema.elements['foo'].type.content.mixed)
+
+    def test_mixed_content_extension__issue_337(self):
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+            <!-- Valid schema: the derived type adds empty content -->
+            <xs:complexType name="baseType" mixed="true">
+                <xs:sequence>
+                    <xs:element name="elem1"/>
+                </xs:sequence>
+            </xs:complexType>
+            <xs:complexType name="derivedType">
+                <xs:complexContent>
+                    <xs:extension base="baseType">
+                    </xs:extension>
+                </xs:complexContent>
+            </xs:complexType>
+        </xs:schema>"""))
+
+        self.assertTrue(schema.types['baseType'].mixed)
+        self.assertEqual(schema.types['baseType'].content_type_label, 'mixed')
+        self.assertTrue(schema.types['derivedType'].mixed)
+        self.assertEqual(schema.types['derivedType'].content_type_label, 'mixed')
+
+        with self.assertRaises(XMLSchemaParseError):
+            self.schema_class(dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+                <!-- Invalid schema: the derived type adds element-only content -->
+                <xs:complexType name="baseType" mixed="true">
+                    <xs:sequence>
+                        <xs:element name="elem1"/>
+                    </xs:sequence>
+                </xs:complexType>
+                <xs:complexType name="derivedType">
+                    <xs:complexContent>
+                        <xs:extension base="baseType">
+                            <xs:sequence>
+                                <xs:element name="elem2"/>
+                            </xs:sequence>
+                        </xs:extension>
+                    </xs:complexContent>
+                </xs:complexType>
+            </xs:schema>"""))
+
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+            <!-- Valid schema: the derived type adds mixed content -->
+            <xs:complexType name="baseType" mixed="true">
+            </xs:complexType>
+            <xs:complexType name="derivedType">
+                <xs:complexContent mixed="true">
+                    <xs:extension base="baseType">
+                        <xs:sequence>
+                            <xs:element name="elem1"/>
+                        </xs:sequence>
+                    </xs:extension>
+                </xs:complexContent>
+            </xs:complexType>
+        </xs:schema>"""))
+
+        self.assertTrue(schema.types['baseType'].mixed)
+        self.assertEqual(schema.types['baseType'].content_type_label, 'mixed')
+        self.assertTrue(schema.types['derivedType'].mixed)
+        self.assertEqual(schema.types['derivedType'].content_type_label, 'mixed')
+
+        with self.assertRaises(XMLSchemaParseError):
+            self.schema_class(dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+                <!-- Invalid schema: the derived type adds element-only content -->
+                <xs:complexType name="baseType" mixed="true">
+                </xs:complexType>
+                <xs:complexType name="derivedType">
+                    <xs:complexContent>
+                        <xs:extension base="baseType">
+                            <xs:sequence>
+                                <xs:element name="elem1"/>
+                            </xs:sequence>
+                        </xs:extension>
+                    </xs:complexContent>
+                </xs:complexType>
+            </xs:schema>"""))
+
+    def test_empty_content_extension(self):
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+            <xs:complexType name="baseType" mixed="false">
+            </xs:complexType>
+            <xs:complexType name="derivedType" mixed="true">
+                <xs:complexContent>
+                    <xs:extension base="baseType"/>
+                </xs:complexContent>
+            </xs:complexType>
+        </xs:schema>"""))
+
+        self.assertFalse(schema.types['baseType'].mixed)
+        self.assertEqual(schema.types['baseType'].content_type_label, 'empty')
+        self.assertTrue(schema.types['derivedType'].mixed)
+        self.assertEqual(schema.types['derivedType'].content_type_label, 'mixed')
+
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+            <xs:complexType name="baseType" mixed="false">
+            </xs:complexType>
+            <xs:complexType name="derivedType">
+                <xs:complexContent>
+                    <xs:extension base="baseType"/>
+                </xs:complexContent>
+            </xs:complexType>
+        </xs:schema>"""))
+
+        self.assertFalse(schema.types['baseType'].mixed)
+        self.assertEqual(schema.types['baseType'].content_type_label, 'empty')
+        self.assertFalse(schema.types['derivedType'].mixed)
+        self.assertEqual(schema.types['derivedType'].content_type_label, 'empty')
+
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+            <xs:complexType name="baseType" mixed="false">
+            </xs:complexType>
+            <xs:complexType name="derivedType">
+                <xs:complexContent>
+                    <xs:extension base="baseType">
+                        <xs:sequence>
+                            <xs:element name="elem1"/>
+                        </xs:sequence>
+                    </xs:extension>
+                </xs:complexContent>
+            </xs:complexType>
+        </xs:schema>"""))
+
+        self.assertFalse(schema.types['baseType'].mixed)
+        self.assertEqual(schema.types['baseType'].content_type_label, 'empty')
+        self.assertFalse(schema.types['derivedType'].mixed)
+        self.assertEqual(schema.types['derivedType'].content_type_label, 'element-only')
+
+    def test_element_only_content_extension(self):
+
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+            <xs:complexType name="baseType" mixed="false">
+                <xs:sequence>
+                    <xs:element name="elem1"/>
+                </xs:sequence>
+            </xs:complexType>
+            <xs:complexType name="derivedType">
+                <xs:complexContent>
+                    <xs:extension base="baseType"/>
+                </xs:complexContent>
+            </xs:complexType>
+        </xs:schema>"""))
+
+        self.assertFalse(schema.types['baseType'].mixed)
+        self.assertEqual(schema.types['baseType'].content_type_label, 'element-only')
+        self.assertFalse(schema.types['derivedType'].mixed)
+        self.assertEqual(schema.types['derivedType'].content_type_label, 'element-only')
+
+        with self.assertRaises(XMLSchemaParseError):
+            self.schema_class(dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+                <!-- Invalid schema: the derived type adds mixed content -->
+                <xs:complexType name="baseType">
+                    <xs:sequence>
+                        <xs:element name="elem1"/>
+                    </xs:sequence>
+                </xs:complexType>
+                <xs:complexType name="derivedType" mixed="true">
+                    <xs:complexContent>
+                        <xs:extension base="baseType">
+                            <xs:sequence>
+                                <xs:element name="elem2"/>
+                            </xs:sequence>
+                        </xs:extension>
+                    </xs:complexContent>
+                </xs:complexType>
+            </xs:schema>"""))
+
+        with self.assertRaises(XMLSchemaParseError):
+            self.schema_class(dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+                <!-- Invalid schema: the derived type adds mixed content -->
+                <xs:complexType name="baseType">
+                    <xs:sequence>
+                        <xs:element name="elem1"/>
+                    </xs:sequence>
+                </xs:complexType>
+                <xs:complexType name="derivedType" mixed="true">
+                    <xs:complexContent>
+                        <xs:extension base="baseType"/>
+                    </xs:complexContent>
+                </xs:complexType>
+            </xs:schema>"""))
+
+    def test_any_type_extension(self):
+        schema = self.schema_class(dedent("""\
+        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+            <xs:complexType name="derivedType">
+                <xs:complexContent>
+                    <xs:extension base="xs:anyType"/>
+                </xs:complexContent>
+            </xs:complexType>
+        </xs:schema>"""))
+
+        self.assertTrue(schema.types['derivedType'].mixed)
+        self.assertEqual(schema.types['derivedType'].content_type_label, 'mixed')
+
+        xsd_source = dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+                <xs:complexType name="derivedType">
+                    <xs:complexContent mixed="true">
+                        <xs:extension base="xs:anyType">
+                            <xs:sequence>
+                                <xs:element name="elem1"/>
+                            </xs:sequence>
+                        </xs:extension>
+                    </xs:complexContent>
+                </xs:complexType>
+            </xs:schema>""")
+
+        if self.schema_class.XSD_VERSION == '1.0':
+            with self.assertRaises(XMLSchemaModelError):
+                self.schema_class(xsd_source)
+        else:
+            schema = self.schema_class(xsd_source)
+            self.assertTrue(schema.types['derivedType'].mixed)
+            self.assertEqual(schema.types['derivedType'].content_type_label, 'mixed')
+
+        with self.assertRaises(XMLSchemaParseError):
+            self.schema_class(dedent("""\
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+                <!-- Invalid schema: derived type content is element-only -->
+                <xs:complexType name="derivedType">
+                    <xs:complexContent>
+                        <xs:extension base="xs:anyType">
+                            <xs:sequence>
+                                <xs:element name="elem1"/>
+                            </xs:sequence>
+                        </xs:extension>
+                    </xs:complexContent>
+                </xs:complexType>
+            </xs:schema>"""))
+
 
 class TestXsd11ComplexType(TestXsdComplexType):
 
@@ -467,11 +761,11 @@ class TestXsd11ComplexType(TestXsdComplexType):
             </xs:complexType>""")
 
         xsd_type = schema.types['intRange']
-        xsd_type.decode(etree_element('a', attrib={'min': '10', 'max': '19'}))
-        self.assertTrue(xsd_type.is_valid(etree_element('a', attrib={'min': '10', 'max': '19'})))
-        self.assertTrue(xsd_type.is_valid(etree_element('a', attrib={'min': '19', 'max': '19'})))
-        self.assertFalse(xsd_type.is_valid(etree_element('a', attrib={'min': '25', 'max': '19'})))
-        self.assertTrue(xsd_type.is_valid(etree_element('a', attrib={'min': '25', 'max': '100'})))
+        xsd_type.decode(Element('a', attrib={'min': '10', 'max': '19'}))
+        self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '10', 'max': '19'})))
+        self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '19', 'max': '19'})))
+        self.assertFalse(xsd_type.is_valid(Element('a', attrib={'min': '25', 'max': '19'})))
+        self.assertTrue(xsd_type.is_valid(Element('a', attrib={'min': '25', 'max': '100'})))
 
     def test_sequence_extension(self):
         schema = self.schema_class("""
