@@ -11,12 +11,36 @@
 """Tests concerning model groups validation"""
 import unittest
 import os.path
+from textwrap import dedent
+from typing import Any, Union, List, Optional
 
 from xmlschema import XMLSchema10, XMLSchema11
-from xmlschema.validators import XsdElement, ModelVisitor, XMLSchemaValidationError
+from xmlschema.exceptions import XMLSchemaValueError
+from xmlschema.validators.exceptions import XMLSchemaValidationError
+from xmlschema.validators.particles import ParticleMixin
+from xmlschema.validators.models import distinguishable_paths, ModelVisitor, \
+    sort_content, iter_collapsed_content
+from xmlschema.validators.groups import XsdGroup
+from xmlschema.validators.elements import XsdElement
 from xmlschema.testing import XsdValidatorTestCase
 
 
+class ModelGroup(XsdGroup):
+    """A subclass for testing XSD models, that disables element parsing and schema bindings."""
+
+    def __init__(self, model: str, min_occurs: int = 1, max_occurs: Optional[int] = 1) -> None:
+        ParticleMixin.__init__(self, min_occurs, max_occurs)
+        if model not in {'sequence', 'choice', 'all'}:
+            raise XMLSchemaValueError("invalid model {!r} for a group".format(model))
+        self._group: List[Union[ParticleMixin, 'ModelGroup']] = []
+        self.model: str = model
+
+    def __repr__(self) -> str:
+        return '%s(model=%r, occurs=%r)' % (self.__class__.__name__, self.model, self.occurs)
+
+    append: Any
+
+
 class TestModelValidation(XsdValidatorTestCase):
 
     TEST_CASES_DIR = os.path.join(os.path.dirname(__file__), '../test_cases')
@@ -78,6 +102,31 @@ class TestModelValidation(XsdValidatorTestCase):
         else:
             self.assertEqual([e for e in model.stop()], expected or [])
 
+    # --- ModelVisitor methods ---
+
+    def test_iter_group(self):
+        group = ModelGroup('sequence', min_occurs=0, max_occurs=0)
+        model = ModelVisitor(group)
+        self.assertListEqual(list(model.items), [])
+
+        group = ModelGroup('choice')
+        group.append(ParticleMixin())
+        group.append(ParticleMixin())
+        group.append(ParticleMixin())
+
+        model = ModelVisitor(group)
+        model.occurs[group[1]] = 1
+        self.assertEqual(list(model.items), group[1:])
+
+        group = ModelGroup('all')
+        group.append(ParticleMixin())
+        group.append(ParticleMixin())
+        group.append(ParticleMixin())
+
+        model = ModelVisitor(group)
+        model.occurs[group[1]] = 1
+        self.assertEqual(list(model.items), group[2:])
+
     # --- Vehicles schema ---
 
     def test_vehicles_model(self):
@@ -164,7 +213,7 @@ class TestModelValidation(XsdValidatorTestCase):
           </xs:choice>
         </xs:group>
         """
-        group = XMLSchema10.meta_schema.groups['simpleDerivation']
+        group = self.schema_class.meta_schema.groups['simpleDerivation']
 
         model = ModelVisitor(group)
         self.check_advance_true(model)     # <restriction> matches
@@ -572,35 +621,35 @@ class TestModelValidation(XsdValidatorTestCase):
                     <xs:choice>
                         <xs:any maxOccurs="2" processContents="lax"/>
                     </xs:choice>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="a2">
                 <xs:complexType>
                     <xs:choice>
                         <xs:any maxOccurs="2" processContents="strict"/>
                     </xs:choice>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="a3">
                 <xs:complexType>
                     <xs:sequence>
                         <xs:any maxOccurs="2" processContents="lax"/>
                     </xs:sequence>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="a4">
                 <xs:complexType>
                     <xs:choice>
                         <xs:element name="b" maxOccurs="2"/>
                     </xs:choice>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="a5">
                 <xs:complexType>
                     <xs:sequence>
                         <xs:element name="b" maxOccurs="2"/>
                     </xs:sequence>
-                </xs:complexType>                
+                </xs:complexType>
             </xs:element>
             <xs:element name="b"/>
         </xs:schema>""")
@@ -747,6 +796,25 @@ class TestModelValidation(XsdValidatorTestCase):
 
         self.assertIsNone(schema.validate('<root><a/><a/><a/></root>'))
 
+    def test_emptiable_all_model(self):
+        schema = self.schema_class(dedent(
+            """<?xml version="1.0" encoding="UTF-8"?>
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+                <xs:element name="root">
+                    <xs:complexType>
+                        <xs:all minOccurs="0">
+                            <xs:element name="a" />
+                            <xs:element name="b" />
+                        </xs:all>
+                    </xs:complexType>
+                </xs:element>
+            </xs:schema>
+            """))
+
+        self.assertIsNone(schema.validate('<root><b/><a/></root>'))
+        self.assertIsNone(schema.validate('<root/>'))
+        self.assertFalse(schema.is_valid('<root><b/></root>'))
+
     #
     # Tests on issues
     def test_issue_086(self):
@@ -918,56 +986,72 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
 
         self.assertListEqual(
-            model.sort_content([('B2', 10), ('B1', 'abc'), ('B3', True)]),
+            sort_content([('B2', 10), ('B1', 'abc'), ('B3', True)], group),
             [('B1', 'abc'), ('B2', 10), ('B3', True)]
         )
         self.assertListEqual(
-            model.sort_content([('B3', True), ('B2', 10), ('B1', 'abc')]),
+            sort_content([('B3', True), ('B2', 10), ('B1', 'abc')], group),
             [('B1', 'abc'), ('B2', 10), ('B3', True)]
         )
         self.assertListEqual(
-            model.sort_content([('B2', 10), ('B4', None), ('B1', 'abc'), ('B3', True)]),
+            sort_content([('B2', 10), ('B4', None), ('B1', 'abc'), ('B3', True)], group),
             [('B1', 'abc'), ('B2', 10), ('B3', True), ('B4', None)]
         )
+
         content = [('B2', 10), ('B4', None), ('B1', 'abc'), (1, 'hello'), ('B3', True)]
         self.assertListEqual(
-            model.sort_content(content),
+            sort_content(content, group),
             [(1, 'hello'), ('B1', 'abc'), ('B2', 10), ('B3', True), ('B4', None)]
         )
+
         content = [
             (2, 'world!'), ('B2', 10), ('B4', None), ('B1', 'abc'), (1, 'hello'), ('B3', True)
         ]
         self.assertListEqual(
-            model.sort_content(content),
+            sort_content(content, group),
             [(1, 'hello'), ('B1', 'abc'), (2, 'world!'), ('B2', 10), ('B3', True), ('B4', None)]
         )
 
+        content = [
+            ('B2', 10), ('B4', None), ('B1', 'abc'), ('B3', True), (6, 'six'),
+            (5, 'five'), (4, 'four'), (2, 'two'), (3, 'three'), (1, 'one')
+        ]
+        self.assertListEqual(
+            sort_content(content, group),
+            [(1, 'one'), ('B1', 'abc'), (2, 'two'), ('B2', 10), (3, 'three'),
+             ('B3', True), (4, 'four'), ('B4', None), (5, 'five'), (6, 'six')]
+        )
+
         # With a dict-type argument
         content = dict([('B2', [10]), ('B1', ['abc']), ('B3', [True])])
         self.assertListEqual(
-            model.sort_content(content), [('B1', 'abc'), ('B2', 10), ('B3', True)]
+            sort_content(content, group), [('B1', 'abc'), ('B2', 10), ('B3', True)]
         )
         content = dict([('B2', [10]), ('B1', ['abc']), ('B3', [True]), (1, 'hello')])
         self.assertListEqual(
-            model.sort_content(content), [(1, 'hello'), ('B1', 'abc'), ('B2', 10), ('B3', True)]
+            sort_content(content, group),
+            [(1, 'hello'), ('B1', 'abc'), ('B2', 10), ('B3', True)]
         )
 
         # With partial content
-        self.assertListEqual(model.sort_content([]), [])
-        self.assertListEqual(model.sort_content([('B1', 'abc')]), [('B1', 'abc')])
-        self.assertListEqual(model.sort_content([('B2', 10)]), [('B2', 10)])
-        self.assertListEqual(model.sort_content([('B3', True)]), [('B3', True)])
+        self.assertListEqual(sort_content([], group), [])
+        self.assertListEqual(sort_content([('B1', 'abc')], group), [('B1', 'abc')])
+        self.assertListEqual(sort_content([('B2', 10)], group), [('B2', 10)])
+        self.assertListEqual(sort_content([('B3', True)], group), [('B3', True)])
         self.assertListEqual(
-            model.sort_content([('B3', True), ('B1', 'abc')]), [('B1', 'abc'), ('B3', True)]
+            sort_content([('B3', True), ('B1', 'abc')], group),
+            [('B1', 'abc'), ('B3', True)]
         )
         self.assertListEqual(
-            model.sort_content([('B2', 10), ('B1', 'abc')]), [('B1', 'abc'), ('B2', 10)]
+            sort_content([('B2', 10), ('B1', 'abc')], group),
+            [('B1', 'abc'), ('B2', 10)]
         )
         self.assertListEqual(
-            model.sort_content([('B3', True), ('B2', 10)]), [('B2', 10), ('B3', True)]
+            sort_content([('B3', True), ('B2', 10)], group),
+            [('B2', 10), ('B3', True)]
         )
 
     def test_iter_collapsed_content_with_optional_elements(self):
@@ -986,19 +1070,26 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
+        model = ModelVisitor(group)
 
         content = [('B3', 10), ('B4', None), ('B5', True), ('B6', 'alpha'), ('B7', 20)]
         model.restart()
         self.assertListEqual(
             list(model.iter_collapsed_content(content)), content
         )
+        self.assertListEqual(
+            list(iter_collapsed_content(content, group)), content
+        )
 
         content = [('B3', 10), ('B5', True), ('B6', 'alpha'), ('B7', 20)]  # Missing B4
         model.restart()
         self.assertListEqual(
             list(model.iter_collapsed_content(content)), content
         )
+        self.assertListEqual(
+            list(iter_collapsed_content(content, group)), content
+        )
 
     def test_iter_collapsed_content_with_repeated_elements(self):
         schema = self.get_schema("""
@@ -1016,27 +1107,19 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
 
         content = [
             ('B3', 10), ('B4', None), ('B5', True), ('B5', False), ('B6', 'alpha'), ('B7', 20)
         ]
-        self.assertListEqual(
-            list(model.iter_collapsed_content(content)), content
-        )
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B3', 10), ('B3', 11), ('B3', 12), ('B4', None), ('B5', True),
                    ('B5', False), ('B6', 'alpha'), ('B7', 20), ('B7', 30)]
-        model.restart()
-        self.assertListEqual(
-            list(model.iter_collapsed_content(content)), content
-        )
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B3', 10), ('B3', 11), ('B3', 12), ('B4', None), ('B5', True), ('B5', False)]
-        model.restart()
-        self.assertListEqual(
-            list(model.iter_collapsed_content(content)), content
-        )
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
     def test_iter_collapsed_content_with_repeated_groups(self):
         schema = self.get_schema("""
@@ -1049,38 +1132,33 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
 
         content = [('B1', 1), ('B1', 2), ('B2', 3), ('B2', 4)]
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)),
+            list(iter_collapsed_content(content, group)),
             [('B1', 1), ('B2', 3), ('B1', 2), ('B2', 4)]
         )
 
         # Model broken by unknown element at start
         content = [('X', None), ('B1', 1), ('B1', 2), ('B2', 3), ('B2', 4)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 1), ('X', None), ('B1', 2), ('B2', 3), ('B2', 4)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 1), ('B1', 2), ('X', None), ('B2', 3), ('B2', 4)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 1), ('B1', 2), ('B2', 3), ('X', None), ('B2', 4)]
-        model.restart()
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)),
+            list(iter_collapsed_content(content, group)),
             [('B1', 1), ('B2', 3), ('B1', 2), ('X', None), ('B2', 4)]
         )
 
         content = [('B1', 1), ('B1', 2), ('B2', 3), ('B2', 4), ('X', None)]
-        model.restart()
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)),
+            list(iter_collapsed_content(content, group)),
             [('B1', 1), ('B2', 3), ('B1', 2), ('B2', 4), ('X', None)]
         )
 
@@ -1096,34 +1174,94 @@ class TestModelBasedSorting(XsdValidatorTestCase):
             </xs:complexType>
             """)
 
-        model = ModelVisitor(schema.types['A_type'].content)
+        group = schema.types['A_type'].content
 
         content = [('B1', 'abc'), ('B2', 10), ('B3', False)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B3', False), ('B1', 'abc'), ('B2', 10)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 'abc'), ('B3', False), ('B2', 10)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('B1', 'abc'), ('B1', 'def'), ('B2', 10), ('B3', False)]
-        model.restart()
         self.assertListEqual(
-            list(model.iter_collapsed_content(content)),
+            list(iter_collapsed_content(content, group)),
             [('B1', 'abc'), ('B2', 10), ('B3', False), ('B1', 'def')]
         )
 
         content = [('B1', 'abc'), ('B2', 10), ('X', None)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
 
         content = [('X', None), ('B1', 'abc'), ('B2', 10), ('B3', False)]
-        model.restart()
-        self.assertListEqual(list(model.iter_collapsed_content(content)), content)
+        self.assertListEqual(list(iter_collapsed_content(content, group)), content)
+
+
+class TestModelPaths(unittest.TestCase):
+
+    def test_distinguishable_paths_one_level(self):
+        group = ModelGroup('sequence', min_occurs=0)
+        group.append(ModelGroup('sequence'))
+        group.append(ModelGroup('sequence'))
+        group[0].append(ParticleMixin())
+        group[1].append(ParticleMixin())
+
+        path1 = [group[0]]
+        path2 = [group[1]]
+        self.assertTrue(distinguishable_paths(path1, path2))  # Disjoined paths
+        self.assertTrue(distinguishable_paths(path1, []))
+
+        with self.assertRaises(IndexError):
+            distinguishable_paths([], path2)  # path1 cannot be empty
+
+        path1 = [group, group[0]]
+        path2 = [group, group[1]]
+        self.assertTrue(distinguishable_paths(path1, path2))
+        group[0].min_occurs = 0
+        self.assertFalse(distinguishable_paths(path1, path2))
+        group.max_occurs = 0
+        self.assertTrue(distinguishable_paths(path1, path2))
+
+    def test_distinguishable_paths_two_levels(self):
+        group = ModelGroup('sequence', min_occurs=0)
+        group.append(ModelGroup('choice'))
+        group.append(ModelGroup('choice'))
+        group[0].append(ParticleMixin())
+        group[0].append(ParticleMixin())
+        group[1].append(ParticleMixin())
+        group[1].append(ParticleMixin())
+
+        path1 = [group, group[0], group[0][0]]
+        path2 = [group, group[1], group[1][0]]
+        self.assertTrue(distinguishable_paths(path1, path2))  # All univocal subgroups
+        group[0].max_occurs = 2
+        self.assertFalse(distinguishable_paths(path1, path2))
+
+        group[0].max_occurs = 1
+        group[0].min_occurs = 0
+        self.assertFalse(distinguishable_paths(path1, path2))
+
+        group.max_occurs = None
+        self.assertFalse(distinguishable_paths(path1, path2))
+
+    def test_distinguishable_paths_three_levels(self):
+        group = ModelGroup('sequence', min_occurs=0)
+        group.append(ModelGroup('choice'))
+        group.append(ModelGroup('choice'))
+        group[0].append(ModelGroup('choice'))
+        group[1].append(ModelGroup('choice'))
+        group[0][0].append(ParticleMixin())
+        group[0][0].append(ParticleMixin())
+        group[1][0].append(ParticleMixin())
+        group[1][0].append(ParticleMixin())
+
+        path1 = [group, group[0], group[0][0], group[0][0][0]]
+        path2 = [group, group[1], group[1][0], group[1][0][0]]
+        self.assertTrue(distinguishable_paths(path1, path2))  # All univocal subgroups
+
+        group[0][0][1].min_occurs = 0
+        self.assertFalse(distinguishable_paths(path1, path2))  # All univocal subgroups
 
 
 if __name__ == '__main__':
