@@ -13,46 +13,62 @@ This module contains declarations and classes for XML Schema constraint facets.
 import re
 import math
 import operator
-from collections.abc import MutableSequence
+from abc import abstractmethod
+from typing import TYPE_CHECKING, cast, Any, List, Optional, Pattern, Union, \
+    MutableSequence, overload, Tuple
+from xml.etree.ElementTree import Element
+
 from elementpath import XPath2Parser, XPathContext, ElementPathError, \
-    translate_pattern, RegexError
+    translate_pattern, RegexError, ElementNode
 
-from ..etree import etree_element
 from ..names import XSD_LENGTH, XSD_MIN_LENGTH, XSD_MAX_LENGTH, XSD_ENUMERATION, \
     XSD_INTEGER, XSD_WHITE_SPACE, XSD_PATTERN, XSD_MAX_INCLUSIVE, XSD_MAX_EXCLUSIVE, \
     XSD_MIN_INCLUSIVE, XSD_MIN_EXCLUSIVE, XSD_TOTAL_DIGITS, XSD_FRACTION_DIGITS, \
-    XSD_ASSERTION, XSD_DECIMAL, XSD_EXPLICIT_TIMEZONE, XSD_NOTATION_TYPE, XSD_QNAME
+    XSD_ASSERTION, XSD_DECIMAL, XSD_EXPLICIT_TIMEZONE, XSD_NOTATION_TYPE, XSD_QNAME, \
+    XSD_ANNOTATION
+from ..aliases import ElementType, SchemaType, AtomicValueType, BaseXsdType
+from ..translation import gettext as _
 from ..helpers import count_digits, local_name
 from .exceptions import XMLSchemaValidationError, XMLSchemaDecodeError
-from .xsdbase import XsdComponent
+from .xsdbase import XsdComponent, XsdAnnotation
+
+if TYPE_CHECKING:
+    from .simple_types import XsdList, XsdAtomicRestriction
+
+LaxDecodeType = Tuple[Any, List[XMLSchemaValidationError]]
 
 
 class XsdFacet(XsdComponent):
     """
     XML Schema constraining facets base class.
     """
+    value: Optional[AtomicValueType]
+    base_type: Optional[BaseXsdType]
+    base_value: Optional[AtomicValueType]
     fixed = False
 
-    def __init__(self, elem, schema, parent, base_type):
+    def __init__(self, elem: ElementType,
+                 schema: SchemaType,
+                 parent: Union['XsdList', 'XsdAtomicRestriction'],
+                 base_type: Optional[BaseXsdType]) -> None:
         self.base_type = base_type
         super(XsdFacet, self).__init__(elem, schema, parent)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return '%s(value=%r, fixed=%r)' % (self.__class__.__name__, self.value, self.fixed)
 
-    def __call__(self, value):
+    def __call__(self, value: Any) -> None:
         try:
             self._validator(value)
         except TypeError:
-            reason = "invalid type {!r} provided".format(type(value))
+            reason = _("invalid type {!r} provided").format(type(value))
             raise XMLSchemaValidationError(self, value, reason) from None
 
     @staticmethod
-    def _validator(_):
+    def _validator(_: Any) -> None:
         return
 
-    def _parse(self):
-        super(XsdFacet, self)._parse()
+    def _parse(self) -> None:
         if 'fixed' in self.elem.attrib and self.elem.attrib['fixed'] in ('true', '1'):
             self.fixed = True
         base_facet = self.base_facet
@@ -66,31 +82,34 @@ class XsdFacet(XsdComponent):
         else:
             if base_facet is not None and base_facet.fixed and \
                     base_facet.value is not None and self.value != base_facet.value:
-                self.parse_error("{!r} facet value is fixed to {!r}"
-                                 .format(local_name(self.elem.tag), base_facet.value))
+                msg = _("{0!r} facet value is fixed to {1!r}")
+                self.parse_error(msg.format(local_name(self.elem.tag), base_facet.value))
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> Union[None, AtomicValueType, Pattern[str]]:
         self.value = elem.attrib['value']  # pragma: no cover
+        return None
 
     @property
-    def built(self):
+    def built(self) -> bool:
         return True  # pragma: no cover
 
     @property
-    def base_facet(self):
+    def base_facet(self) -> Optional['XsdFacet']:
         """
         An object of the same type if the instance has a base facet, `None` otherwise.
         """
-        base_type = self.base_type
+        base_type: Optional[BaseXsdType] = self.base_type
         tag = self.elem.tag
         while True:
+            if base_type is None:
+                return None
             try:
-                return base_type.facets[tag]
+                base_facet = base_type.facets[tag]  # type: ignore[union-attr]
             except (AttributeError, KeyError):
-                if hasattr(base_type, 'base_type'):
-                    base_type = base_type.base_type
-                else:
-                    return None
+                base_type = base_type.base_type
+            else:
+                assert isinstance(base_facet, self.__class__)
+                return base_facet
 
 
 class XsdWhiteSpaceFacet(XsdFacet):
@@ -105,31 +124,32 @@ class XsdWhiteSpaceFacet(XsdFacet):
           Content: (annotation?)
         </whiteSpace>
     """
+    value: str
     _ADMITTED_TAGS = XSD_WHITE_SPACE,
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> None:
         self.value = elem.attrib['value']
         if self.value == 'collapse':
-            self._validator = self.collapse_white_space_validator
+            self._validator = self.collapse_white_space_validator  # type: ignore[assignment]
         elif self.value == 'replace':
             if self.base_value == 'collapse':
-                self.parse_error("facet value can be only 'collapse'")
-            self._validator = self.replace_white_space_validator
+                self.parse_error(_("facet value can be only 'collapse'"))
+            self._validator = self.replace_white_space_validator  # type: ignore[assignment]
         elif self.base_value == 'collapse':
-            self.parse_error("facet value can be only 'collapse'")
+            self.parse_error(_("facet value can be only 'collapse'"))
         elif self.base_value == 'replace':
-            self.parse_error("facet value can be only 'replace' or 'collapse'")
+            self.parse_error(_("facet value can be only 'replace' or 'collapse'"))
 
-    def replace_white_space_validator(self, value):
+    def replace_white_space_validator(self, value: str) -> None:
         if '\t' in value or '\n' in value:
             raise XMLSchemaValidationError(
-                self, value, "value contains tabs or newlines"
+                self, value, _("value contains tabs or newlines")
             )
 
-    def collapse_white_space_validator(self, value):
+    def collapse_white_space_validator(self, value: str) -> None:
         if '\t' in value or '\n' in value or '  ' in value:
             raise XMLSchemaValidationError(
-                self, value, "value contains non collapsed white spaces"
+                self, value, _("value contains non collapsed white spaces")
             )
 
 
@@ -145,21 +165,25 @@ class XsdLengthFacet(XsdFacet):
           Content: (annotation?)
         </length>
     """
+    value: int
+    base_type: BaseXsdType
+    base_value: Optional[int]
     _ADMITTED_TAGS = XSD_LENGTH,
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> None:
         self.value = int(elem.attrib['value'])
         if self.base_value is not None and self.value != self.base_value:
-            self.parse_error("base facet has a different length ({})".format(self.base_value))
+            msg = _("base facet has a different length ({})")
+            self.parse_error(msg.format(self.base_value))
 
         primitive_type = getattr(self.base_type, 'primitive_type', None)
         if primitive_type is None or primitive_type.name not in {XSD_QNAME, XSD_NOTATION_TYPE}:
             # See: https://www.w3.org/Bugs/Public/show_bug.cgi?id=4009
-            self._validator = self._length_validator
+            self._validator = self._length_validator  # type: ignore[assignment]
 
-    def _length_validator(self, value):
+    def _length_validator(self, value: Any) -> None:
         if len(value) != self.value:
-            reason = "length has to be {!r}".format(self.value)
+            reason = _("length has to be {!r}").format(self.value)
             raise XMLSchemaValidationError(self, value, reason)
 
 
@@ -175,21 +199,25 @@ class XsdMinLengthFacet(XsdFacet):
           Content: (annotation?)
         </minLength>
     """
+    value: int
+    base_type: BaseXsdType
+    base_value: Optional[int]
     _ADMITTED_TAGS = XSD_MIN_LENGTH,
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> None:
         self.value = int(elem.attrib['value'])
         if self.base_value is not None and self.value < self.base_value:
-            self.parse_error("base facet has a greater min length ({})".format(self.base_value))
+            msg = _("base facet has a greater min length ({})")
+            self.parse_error(msg.format(self.base_value))
 
         primitive_type = getattr(self.base_type, 'primitive_type', None)
         if primitive_type is None or primitive_type.name not in {XSD_QNAME, XSD_NOTATION_TYPE}:
             # See: https://www.w3.org/Bugs/Public/show_bug.cgi?id=4009
-            self._validator = self._min_length_validator
+            self._validator = self._min_length_validator  # type: ignore[assignment]
 
-    def _min_length_validator(self, value):
+    def _min_length_validator(self, value: Any) -> None:
         if len(value) < self.value:
-            reason = "value length cannot be lesser than {!r}".format(self.value)
+            reason = _("value length cannot be lesser than {!r}").format(self.value)
             raise XMLSchemaValidationError(self, value, reason)
 
 
@@ -205,21 +233,25 @@ class XsdMaxLengthFacet(XsdFacet):
           Content: (annotation?)
         </maxLength>
     """
+    value: int
+    base_type: BaseXsdType
+    base_value: Optional[int]
     _ADMITTED_TAGS = XSD_MAX_LENGTH,
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> None:
         self.value = int(elem.attrib['value'])
         if self.base_value is not None and self.value > self.base_value:
-            self.parse_error("base type has a lesser max length ({})".format(self.base_value))
+            msg = _("base type has a lesser max length ({})")
+            self.parse_error(msg.format(self.base_value))
 
         primitive_type = getattr(self.base_type, 'primitive_type', None)
         if primitive_type is None or primitive_type.name not in {XSD_QNAME, XSD_NOTATION_TYPE}:
             # See: https://www.w3.org/Bugs/Public/show_bug.cgi?id=4009
-            self._validator = self._min_length_validator
+            self._validator = self._max_length_validator  # type: ignore[assignment]
 
-    def _min_length_validator(self, value):
+    def _max_length_validator(self, value: Any) -> None:
         if len(value) > self.value:
-            reason = "value length cannot be greater than {!r}".format(self.value)
+            reason = _("value length cannot be greater than {!r}").format(self.value)
             raise XMLSchemaValidationError(self, value, reason)
 
 
@@ -235,17 +267,19 @@ class XsdMinInclusiveFacet(XsdFacet):
           Content: (annotation?)
         </minInclusive>
     """
+    base_type: BaseXsdType
     _ADMITTED_TAGS = XSD_MIN_INCLUSIVE,
 
-    def _parse_value(self, elem):
-        self.value, errors = self.base_type.decode(elem.attrib['value'], validation='lax')
+    def _parse_value(self, elem: ElementType) -> None:
+        value = elem.attrib['value']
+        self.value, errors = cast(LaxDecodeType, self.base_type.decode(value, 'lax'))
         for e in errors:
-            self.parse_error("invalid restriction: {}".format(e.reason))
+            self.parse_error(_("invalid restriction: {}").format(e.reason))
 
-    def __call__(self, value):
+    def __call__(self, value: Any) -> None:
         try:
             if value < self.value:
-                reason = "value has to be greater or equal than {!r}".format(self.value)
+                reason = _("value has to be greater or equal than {!r}").format(self.value)
                 raise XMLSchemaValidationError(self, value, reason)
         except TypeError as err:
             raise XMLSchemaValidationError(self, value, str(err)) from None
@@ -263,22 +297,25 @@ class XsdMinExclusiveFacet(XsdFacet):
           Content: (annotation?)
         </minExclusive>
     """
+    base_type: BaseXsdType
     _ADMITTED_TAGS = XSD_MIN_EXCLUSIVE,
 
-    def _parse_value(self, elem):
-        self.value, errors = self.base_type.decode(elem.attrib['value'], validation='lax')
+    def _parse_value(self, elem: ElementType) -> None:
+        value = elem.attrib['value']
+        self.value, errors = cast(LaxDecodeType, self.base_type.decode(value, 'lax'))
         for e in errors:
             if not isinstance(e.validator, self.__class__) or e.validator.value != self.value:
-                self.parse_error("invalid restriction: {}".format(e.reason))
+                self.parse_error(_("invalid restriction: {}").format(e.reason))
 
-        facet = self.base_type.get_facet(XSD_MAX_INCLUSIVE)
+        facet: Any = self.base_type.get_facet(XSD_MAX_INCLUSIVE)
         if facet is not None and facet.value == self.value:
-            self.parse_error("invalid restriction: {} is also the maximum".format(self.value))
+            msg = _("invalid restriction: {} is also the maximum")
+            self.parse_error(msg.format(self.value))
 
-    def __call__(self, value):
+    def __call__(self, value: Any) -> None:
         try:
             if value <= self.value:
-                reason = "value has to be greater than {!r}".format(self.value)
+                reason = _("value has to be greater than {!r}").format(self.value)
                 raise XMLSchemaValidationError(self, value, reason)
         except TypeError as err:
             raise XMLSchemaValidationError(self, value, str(err)) from None
@@ -296,17 +333,19 @@ class XsdMaxInclusiveFacet(XsdFacet):
           Content: (annotation?)
         </maxInclusive>
     """
+    base_type: BaseXsdType
     _ADMITTED_TAGS = XSD_MAX_INCLUSIVE,
 
-    def _parse_value(self, elem):
-        self.value, errors = self.base_type.decode(elem.attrib['value'], validation='lax')
+    def _parse_value(self, elem: ElementType) -> None:
+        value = elem.attrib['value']
+        self.value, errors = cast(LaxDecodeType, self.base_type.decode(value, 'lax'))
         for e in errors:
-            self.parse_error("invalid restriction: {}".format(e.reason))
+            self.parse_error(_("invalid restriction: {}").format(e.reason))
 
-    def __call__(self, value):
+    def __call__(self, value: Any) -> None:
         try:
             if value > self.value:
-                reason = "value has to be lesser or equal than {!r}".format(self.value)
+                reason = _("value has to be less than or equal than {!r}").format(self.value)
                 raise XMLSchemaValidationError(self, value, reason)
         except TypeError as err:
             raise XMLSchemaValidationError(self, value, str(err)) from None
@@ -324,22 +363,25 @@ class XsdMaxExclusiveFacet(XsdFacet):
           Content: (annotation?)
         </maxExclusive>
     """
+    base_type: BaseXsdType
     _ADMITTED_TAGS = XSD_MAX_EXCLUSIVE,
 
-    def _parse_value(self, elem):
-        self.value, errors = self.base_type.decode(elem.attrib['value'], validation='lax')
+    def _parse_value(self, elem: ElementType) -> None:
+        value = elem.attrib['value']
+        self.value, errors = cast(LaxDecodeType, self.base_type.decode(value, 'lax'))
         for e in errors:
             if not isinstance(e.validator, self.__class__) or e.validator.value != self.value:
-                self.parse_error("invalid restriction: {}".format(e.reason))
+                self.parse_error(_("invalid restriction: {}").format(e.reason))
 
-        facet = self.base_type.get_facet(XSD_MIN_INCLUSIVE)
+        facet: Any = self.base_type.get_facet(XSD_MIN_INCLUSIVE)
         if facet is not None and facet.value == self.value:
-            self.parse_error("invalid restriction: {} is also the minimum".format(self.value))
+            msg = _("invalid restriction: {} is also the minimum")
+            self.parse_error(msg.format(self.value))
 
-    def __call__(self, value):
+    def __call__(self, value: Any) -> None:
         try:
             if value >= self.value:
-                reason = "value has to be lesser than {!r}".format(self.value)
+                reason = _("value has to be lesser than {!r}").format(self.value)
                 raise XMLSchemaValidationError(self, value, reason)
         except TypeError as err:
             raise XMLSchemaValidationError(self, value, str(err)) from None
@@ -357,9 +399,11 @@ class XsdTotalDigitsFacet(XsdFacet):
           Content: (annotation?)
         </totalDigits>
     """
+    value: int
+    base_type: BaseXsdType
     _ADMITTED_TAGS = XSD_TOTAL_DIGITS,
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> None:
         # Errors are detected by meta-schema validation. For schemas with
         # 'lax' validation mode use 9999 in case of an invalid value.
         try:
@@ -370,21 +414,20 @@ class XsdTotalDigitsFacet(XsdFacet):
             if self.value < 1:
                 self.value = 9999
 
-            facet = self.base_type.get_facet(XSD_TOTAL_DIGITS)
+            facet: Any = self.base_type.get_facet(XSD_TOTAL_DIGITS)
             if facet is not None and facet.value < self.value:
-                self.parse_error(
-                    "invalid restriction: base value is lower ({})".format(facet.value)
-                )
+                msg = _("invalid restriction: base value is lower ({})")
+                self.parse_error(msg.format(facet.value))
 
-    def __call__(self, value):
+    def __call__(self, value: Any) -> None:
         try:
             if operator.add(*count_digits(value)) <= self.value:
                 return
         except (TypeError, ValueError, ArithmeticError) as err:
             raise XMLSchemaValidationError(self, value, str(err)) from None
         else:
-            reason = "the number of digits has to be lesser or equal " \
-                     "than {!r}".format(self.value)
+            reason = _("the number of digits has to be lesser or equal "
+                       "than {!r}").format(self.value)
             raise XMLSchemaValidationError(self, value, reason)
 
 
@@ -400,16 +443,21 @@ class XsdFractionDigitsFacet(XsdFacet):
           Content: (annotation?)
         </fractionDigits>
     """
+    value: int
+    base_type: BaseXsdType
     _ADMITTED_TAGS = XSD_FRACTION_DIGITS,
 
-    def __init__(self, elem, schema, parent, base_type):
+    def __init__(self, elem: ElementType,
+                 schema: SchemaType,
+                 parent: 'XsdAtomicRestriction',
+                 base_type: BaseXsdType) -> None:
+
         super(XsdFractionDigitsFacet, self).__init__(elem, schema, parent, base_type)
         if not base_type.is_derived(self.maps.types[XSD_DECIMAL]):
-            self.parse_error(
-                "fractionDigits facet can be applied only to types derived from xs:decimal"
-            )
+            msg = _("fractionDigits facet can be applied only to types derived from xs:decimal")
+            self.parse_error(msg)
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> None:
         # Errors are detected by meta-schema validation. For schemas with
         # 'lax' validation mode use 9999 in case of an invalid value.
         try:
@@ -420,24 +468,23 @@ class XsdFractionDigitsFacet(XsdFacet):
             if self.value < 0:
                 self.value = 9999
             elif self.value > 0 and self.base_type.is_derived(self.maps.types[XSD_INTEGER]):
-                raise ValueError("fractionDigits facet value has to be 0 "
-                                 "for types derived from xs:integer.")
+                msg = _("fractionDigits facet value must be 0 for types derived from xs:integer")
+                raise ValueError(msg)
 
-            facet = self.base_type.get_facet(XSD_FRACTION_DIGITS)
+            facet: Any = self.base_type.get_facet(XSD_FRACTION_DIGITS)
             if facet is not None and facet.value < self.value:
-                self.parse_error(
-                    "invalid restriction: base value is lower ({})".format(facet.value)
-                )
+                msg = _("invalid restriction: base value is lower ({})")
+                self.parse_error(msg.format(facet.value))
 
-    def __call__(self, value):
+    def __call__(self, value: Any) -> None:
         try:
             if count_digits(value)[1] <= self.value:
                 return
         except (TypeError, ValueError, ArithmeticError) as err:
             raise XMLSchemaValidationError(self, value, str(err)) from None
         else:
-            reason = "the number of fraction digits has to be lesser " \
-                     "or equal than {!r}".format(self.value)
+            reason = _("the number of fraction digits has to be lesser "
+                       "or equal than {!r}").format(self.value)
             raise XMLSchemaValidationError(self, value, reason)
 
 
@@ -453,35 +500,36 @@ class XsdExplicitTimezoneFacet(XsdFacet):
           Content: (annotation?)
         </explicitTimezone>
     """
+    value: str
+    base_type: BaseXsdType
     _ADMITTED_TAGS = XSD_EXPLICIT_TIMEZONE,
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> None:
         self.value = elem.attrib['value']
         if self.value == 'prohibited':
-            self._validator = self._prohibited_timezone_validator
+            self._validator = self._prohibited_timezone_validator  # type: ignore[assignment]
         elif self.value == 'required':
-            self._validator = self._required_timezone_validator
+            self._validator = self._required_timezone_validator  # type: ignore[assignment]
         elif self.value != 'optional':
             self.value = 'optional'  # Error already detected by meta-schema validation
 
-        facet = self.base_type.get_facet(XSD_EXPLICIT_TIMEZONE)
+        facet: Any = self.base_type.get_facet(XSD_EXPLICIT_TIMEZONE)
         if facet is not None and facet.value != self.value and facet.value != 'optional':
-            self.parse_error("invalid restriction from {!r}".format(facet.value))
+            msg = _("invalid restriction from {!r}")
+            self.parse_error(msg.format(facet.value))
 
-    def _required_timezone_validator(self, value):
+    def _required_timezone_validator(self, value: Any) -> None:
         if value.tzinfo is None:
-            raise XMLSchemaValidationError(
-                self, value, "time zone required for value {!r}".format(self.value)
-            )
+            reason = _("time zone required for value {!r}").format(self.value)
+            raise XMLSchemaValidationError(self, value, reason)
 
-    def _prohibited_timezone_validator(self, value):
+    def _prohibited_timezone_validator(self, value: Any) -> None:
         if value.tzinfo is not None:
-            raise XMLSchemaValidationError(
-                self, value, "time zone prohibited for value {!r}".format(self.value)
-            )
+            reason = _("time zone prohibited for value {!r}").format(self.value)
+            raise XMLSchemaValidationError(self, value, reason)
 
 
-class XsdEnumerationFacets(MutableSequence, XsdFacet):
+class XsdEnumerationFacets(MutableSequence[ElementType], XsdFacet):
     """
     Sequence of XSD *enumeration* facets. Values are validates if match any of enumeration values.
 
@@ -492,17 +540,20 @@ class XsdEnumerationFacets(MutableSequence, XsdFacet):
           Content: (annotation?)
         </enumeration>
     """
+    base_type: BaseXsdType
     _ADMITTED_TAGS = {XSD_ENUMERATION}
 
-    def __init__(self, elem, schema, parent, base_type):
+    def __init__(self, elem: ElementType,
+                 schema: SchemaType,
+                 parent: 'XsdAtomicRestriction',
+                 base_type: BaseXsdType) -> None:
         XsdFacet.__init__(self, elem, schema, parent, base_type)
 
-    def _parse(self):
-        super(XsdFacet, self)._parse()
+    def _parse(self) -> None:
         self._elements = [self.elem]
         self.enumeration = [self._parse_value(self.elem)]
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> Optional[AtomicValueType]:
         try:
             value = self.base_type.decode(elem.attrib['value'], namespaces=self.schema.namespaces)
         except KeyError:
@@ -511,36 +562,49 @@ class XsdEnumerationFacets(MutableSequence, XsdFacet):
             self.parse_error(err, elem)
         else:
             if self.base_type.name == XSD_NOTATION_TYPE:
+                assert isinstance(value, str)
                 try:
                     notation_qname = self.schema.resolve_qname(value)
                 except (KeyError, ValueError, RuntimeError) as err:
                     self.parse_error(err, elem)
                 else:
                     if notation_qname not in self.maps.notations:
-                        msg = "value {!r} must match a notation declaration"
+                        msg = _("value {!r} must match a notation declaration")
                         self.parse_error(msg.format(value), elem)
-            return value
+            return cast(AtomicValueType, value)
+        return None
 
-    # Implements the abstract methods of MutableSequence
-    def __getitem__(self, i):
+    @overload
+    @abstractmethod
+    def __getitem__(self, i: int) -> ElementType: ...
+
+    @overload
+    @abstractmethod
+    def __getitem__(self, s: slice) -> MutableSequence[ElementType]: ...
+
+    def __getitem__(self, i: Union[int, slice]) \
+            -> Union[ElementType, MutableSequence[ElementType]]:
         return self._elements[i]
 
-    def __setitem__(self, i, elem):
-        self._elements[i] = elem
-        self.enumeration[i] = self._parse_value(elem)
+    def __setitem__(self, i: Union[int, slice], o: Any) -> None:
+        self._elements[i] = o
+        if isinstance(i, int):
+            self.enumeration[i] = self._parse_value(o)
+        else:
+            self.enumeration[i] = [self._parse_value(e) for e in o]
 
-    def __delitem__(self, i):
+    def __delitem__(self, i: Union[int, slice]) -> None:
         del self._elements[i]
         del self.enumeration[i]
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self._elements)
 
-    def insert(self, i, elem):
+    def insert(self, i: int, elem: ElementType) -> None:
         self._elements.insert(i, elem)
         self.enumeration.insert(i, self._parse_value(elem))
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         if len(self.enumeration) > 5:
             return '%s(%s)' % (
                 self.__class__.__name__, '[%s, ...]' % ', '.join(map(repr, self.enumeration[:5]))
@@ -548,24 +612,38 @@ class XsdEnumerationFacets(MutableSequence, XsdFacet):
         else:
             return '%s(%r)' % (self.__class__.__name__, self.enumeration)
 
-    def __call__(self, value):
+    def __call__(self, value: Any) -> None:
         if value in self.enumeration:
             return
+
         try:
             if math.isnan(value):
-                if any(math.isnan(x) for x in self.enumeration):
+                if any(math.isnan(x) for x in self.enumeration):  # type: ignore[arg-type]
+                    return
+            elif math.isinf(value):
+                if any(math.isinf(x) and str(value) == str(x)  # type: ignore[arg-type]
+                       for x in self.enumeration):  # pragma: no cover
                     return
-            elif math.isinf(value) and any(math.isinf(x) and str(value) == str(x)
-                                           for x in self.enumeration):  # pragma: no cover
-                return
         except TypeError:
             pass
 
-        reason = "value must be one of {!r}".format(self.enumeration)
+        reason = _("value must be one of {!r}").format(self.enumeration)
         raise XMLSchemaValidationError(self, value, reason)
 
+    def get_annotation(self, i: int) -> Optional[XsdAnnotation]:
+        """
+        Get the XSD annotation of the i-th enumeration facet.
 
-class XsdPatternFacets(MutableSequence, XsdFacet):
+        :param i: an integer index.
+        :returns: an XsdAnnotation object or `None`.
+        """
+        for child in self._elements[i]:
+            if child.tag == XSD_ANNOTATION:
+                return XsdAnnotation(child, self.schema, self)
+        return None
+
+
+class XsdPatternFacets(MutableSequence[ElementType], XsdFacet):
     """
     Sequence of XSD *pattern* facets. Values are validates if match any of patterns.
 
@@ -577,16 +655,19 @@ class XsdPatternFacets(MutableSequence, XsdFacet):
         </pattern>
     """
     _ADMITTED_TAGS = {XSD_PATTERN}
+    patterns: List[Pattern[str]]
 
-    def __init__(self, elem, schema, parent, base_type):
+    def __init__(self, elem: ElementType,
+                 schema: SchemaType,
+                 parent: 'XsdAtomicRestriction',
+                 base_type: Optional[BaseXsdType]) -> None:
         XsdFacet.__init__(self, elem, schema, parent, base_type)
 
-    def _parse(self):
-        super(XsdFacet, self)._parse()
+    def _parse(self) -> None:
         self._elements = [self.elem]
         self.patterns = [self._parse_value(self.elem)]
 
-    def _parse_value(self, elem):
+    def _parse_value(self, elem: ElementType) -> Pattern[str]:
         try:
             python_pattern = translate_pattern(
                 pattern=elem.attrib['value'],
@@ -602,43 +683,66 @@ class XsdPatternFacets(MutableSequence, XsdFacet):
             self.parse_error(str(err), elem)
             return re.compile(r'^.*$')
 
-    # Implements the abstract methods of MutableSequence
-    def __getitem__(self, i):
+    @overload
+    @abstractmethod
+    def __getitem__(self, i: int) -> ElementType: ...
+
+    @overload
+    @abstractmethod
+    def __getitem__(self, s: slice) -> MutableSequence[ElementType]: ...
+
+    def __getitem__(self, i: Union[int, slice]) \
+            -> Union[ElementType, MutableSequence[ElementType]]:
         return self._elements[i]
 
-    def __setitem__(self, i, elem):
-        self._elements[i] = elem
-        self.patterns[i] = self._parse_value(elem)
+    def __setitem__(self, i: Union[int, slice], o: Any) -> None:
+        self._elements[i] = o
+        if isinstance(i, int):
+            self.patterns[i] = self._parse_value(o)
+        else:
+            self.patterns[i] = [self._parse_value(e) for e in o]
 
-    def __delitem__(self, i):
+    def __delitem__(self, i: Union[int, slice]) -> None:
         del self._elements[i]
         del self.patterns[i]
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self._elements)
 
-    def insert(self, i, elem):
+    def insert(self, i: int, elem: ElementType) -> None:
         self._elements.insert(i, elem)
         self.patterns.insert(i, self._parse_value(elem))
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         s = repr(self.regexps)
         if len(s) < 70:
             return '%s(%s)' % (self.__class__.__name__, s)
         else:
             return '%s(%s...\'])' % (self.__class__.__name__, s[:70])
 
-    def __call__(self, text):
+    def __call__(self, text: str) -> None:
         try:
             if all(pattern.match(text) is None for pattern in self.patterns):
-                reason = "value doesn't match any pattern of {!r}".format(self.regexps)
+                reason = _("value doesn't match any pattern of {!r}").format(self.regexps)
                 raise XMLSchemaValidationError(self, text, reason)
         except TypeError as err:
             raise XMLSchemaValidationError(self, text, str(err)) from None
 
     @property
-    def regexps(self):
-        return [e.get('value', '') for e in self._elements]
+    def regexps(self) -> List[str]:
+        return [e.attrib.get('value', '') for e in self._elements]
+
+    def get_annotation(self, i: int) -> Optional[XsdAnnotation]:
+        """
+        Get the XSD annotation of the i-th pattern facet.
+
+        :param i: an integer index.
+        :returns: an XsdAnnotation object or `None`.
+        """
+        for child in self._elements[i]:
+            if child.tag == XSD_ANNOTATION:
+                return XsdAnnotation(child, self.schema, self)
+        return None
 
 
 class XsdAssertionXPathParser(XPath2Parser):
@@ -650,14 +754,16 @@ XsdAssertionXPathParser.unregister('position')
 
 
 # noinspection PyUnusedLocal
-@XsdAssertionXPathParser.method(XsdAssertionXPathParser.function('last', nargs=0))
-def evaluate(self, context=None):
+@XsdAssertionXPathParser.method(  # type: ignore[no-untyped-def]
+    XsdAssertionXPathParser.function('last', nargs=0))
+def evaluate_last(self, context=None):
     raise self.missing_context("context item size is undefined")
 
 
 # noinspection PyUnusedLocal
-@XsdAssertionXPathParser.method(XsdAssertionXPathParser.function('position', nargs=0))
-def evaluate(self, context=None):
+@XsdAssertionXPathParser.method(  # type: ignore[no-untyped-def]
+    XsdAssertionXPathParser.function('position', nargs=0))
+def evaluate_position(self, context=None):
     raise self.missing_context("context item position is undefined")
 
 
@@ -674,23 +780,22 @@ class XsdAssertionFacet(XsdFacet):
         </assertion>
     """
     _ADMITTED_TAGS = {XSD_ASSERTION}
-    _root = etree_element('root')
+    _root = ElementNode(elem=Element('root'))
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return '%s(test=%r)' % (self.__class__.__name__, self.path)
 
-    def _parse(self):
-        super(XsdFacet, self)._parse()
+    def _parse(self) -> None:
         try:
             self.path = self.elem.attrib['test']
         except KeyError:
-            self.parse_error("missing attribute 'test'")
+            self.parse_error(_("missing attribute 'test'"))
             self.path = 'true()'
 
         try:
-            variable_types = {'value': self.base_type.primitive_type.prefixed_name}
+            value = self.base_type.primitive_type.prefixed_name  # type: ignore[union-attr]
         except AttributeError:
-            variable_types = {'value': self.any_simple_type.prefixed_name}
+            value = self.any_simple_type.prefixed_name
 
         if 'xpathDefaultNamespace' in self.elem.attrib:
             self.xpath_default_namespace = self._parse_xpath_default_namespace(self.elem)
@@ -698,7 +803,9 @@ class XsdAssertionFacet(XsdFacet):
             self.xpath_default_namespace = self.schema.xpath_default_namespace
 
         self.parser = XsdAssertionXPathParser(
-            self.namespaces, strict=False, variable_types=variable_types,
+            namespaces=self.namespaces,
+            strict=False,
+            variable_types={'value': value},
             default_namespace=self.xpath_default_namespace
         )
 
@@ -708,11 +815,11 @@ class XsdAssertionFacet(XsdFacet):
             self.parse_error(err)
             self.token = self.parser.parse('true()')
 
-    def __call__(self, value):
+    def __call__(self, value: AtomicValueType) -> None:
         context = XPathContext(self._root, variables={'value': value})
         try:
             if not self.token.evaluate(context):
-                reason = "value is not true with test path {!r}".format(self.path)
+                reason = _("value is not true with test path {!r}").format(self.path)
                 raise XMLSchemaValidationError(self, value, reason)
         except ElementPathError as err:
             raise XMLSchemaValidationError(self, value, reason=str(err)) from None
@@ -748,3 +855,9 @@ XSD_11_LIST_FACETS = XSD_10_LIST_FACETS | {XSD_ASSERTION}
 
 XSD_10_UNION_FACETS = {XSD_PATTERN, XSD_ENUMERATION}
 XSD_11_UNION_FACETS = MULTIPLE_FACETS = {XSD_PATTERN, XSD_ENUMERATION, XSD_ASSERTION}
+
+
+XsdFacetType = Union[XsdLengthFacet, XsdMinLengthFacet, XsdMaxLengthFacet,
+                     XsdMinInclusiveFacet, XsdMinExclusiveFacet, XsdMaxInclusiveFacet,
+                     XsdMaxExclusiveFacet, XsdTotalDigitsFacet, XsdFractionDigitsFacet,
+                     XsdEnumerationFacets, XsdPatternFacets]
